#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the loss model of pelicun."""

from __future__ import annotations

import re
from copy import deepcopy
from itertools import product
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest

from pelicun import file_io, model, uq
from pelicun.base import ensure_value
from pelicun.model.loss_model import (
    LossModel,
    RepairModel_DS,
    RepairModel_LF,
    _is_for_ds_model,
    _is_for_lf_model,
)
from pelicun.pelicun_warnings import PelicunWarning
from pelicun.tests.basic.test_pelicun_model import TestPelicunModel

if TYPE_CHECKING:
    from pelicun.assessment import Assessment
    from pelicun.model.asset_model import AssetModel


class TestLossModel(TestPelicunModel):
    @pytest.fixture
    def loss_model(self, assessment_instance: Assessment) -> LossModel:
        return deepcopy(assessment_instance.loss)

    @pytest.fixture
    def asset_model_empty(self, assessment_instance: Assessment) -> AssetModel:
        return deepcopy(assessment_instance.asset)

    @pytest.fixture
    def asset_model_a(self, asset_model_empty: AssetModel) -> AssetModel:
        asset = deepcopy(asset_model_empty)
        asset.cmp_marginal_params = pd.DataFrame(
            {
                ('Theta_0'): [1.0, 1.0, 1.0],
                ('Blocks'): [1, 1, 1],
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ('cmp.A', '1', '1', '0'),
                    ('cmp.B', '1', '1', '0'),
                    ('cmp.C', '1', '1', '0'),
                ]
            ),
        ).rename_axis(index=['cmp', 'loc', 'dir', 'uid'])
        asset.generate_cmp_sample(sample_size=10)
        return asset

    @pytest.fixture
    def loss_model_with_ones(self, assessment_instance: Assessment) -> LossModel:
        loss_model = assessment_instance.loss

        # add artificial values to the samples
        data_ds = {}
        for (
            decision_variable,
            consequence,
            component,
            damage_state,
            location,
            direction,
            uid,
        ) in product(
            ('Cost', 'Carbon'),
            ('cmp.A.consequence', 'cmp.B.consequence'),
            ('cmp.A', 'cmp.B'),
            ('DS1', 'DS2'),
            ('1', '2'),  # loc
            ('1', '2'),  # dir
            ('uid1', 'uid2'),
        ):
            data_ds[
                decision_variable,
                consequence,
                component,
                damage_state,
                location,
                direction,
                uid,
            ] = [1.00, 1.00, 1.00]
        loss_model.ds_model.sample = pd.DataFrame(data_ds).rename_axis(
            columns=['dv', 'loss', 'dmg', 'ds', 'loc', 'dir', 'uid']
        )
        data_lf = {}
        for (
            decision_variable,
            consequence,
            component,
            location,
            direction,
            uid,
        ) in product(
            ('Cost', 'Carbon'),
            ('cmp.A.consequence', 'cmp.B.consequence'),
            ('cmp.A', 'cmp.B'),
            ('1', '2'),  # loc
            ('1', '2'),  # dir
            ('uid1', 'uid2'),
        ):
            data_lf[
                decision_variable,
                consequence,
                component,
                location,
                direction,
                uid,
            ] = [1.00, 1.00, 1.00]
        loss_model.lf_model.sample = pd.DataFrame(data_lf).rename_axis(
            columns=['dv', 'loss', 'dmg', 'loc', 'dir', 'uid']
        )

        return loss_model

    def test___init__(self, loss_model: LossModel) -> None:
        assert loss_model.log
        assert loss_model.ds_model
        with pytest.raises(AttributeError):
            loss_model.xyz = 123  # type: ignore

        assert loss_model.ds_model.loss_params is None
        assert loss_model.ds_model.sample is None

        assert len(loss_model._loss_models) == 2

    def test_decision_variables(self, loss_model: LossModel) -> None:
        dvs = ('Cost', 'Time')
        assert loss_model.decision_variables == dvs
        assert loss_model.ds_model.decision_variables == dvs
        assert loss_model.lf_model.decision_variables == dvs

    def test_add_loss_map(
        self, loss_model: LossModel, asset_model_a: AssetModel
    ) -> None:
        loss_model._asmnt.asset = asset_model_a

        loss_map = pd.DataFrame(
            {
                'Repair': ['consequence.A', 'consequence.B'],
            },
            index=['cmp.A', 'cmp.B'],
        )
        loss_model.add_loss_map(loss_map)
        pd.testing.assert_frame_equal(ensure_value(loss_model._loss_map), loss_map)
        for contained_model in loss_model._loss_models:
            pd.testing.assert_frame_equal(
                ensure_value(contained_model.loss_map), loss_map
            )

    def test_load_model_parameters(
        self, loss_model: LossModel, asset_model_a: AssetModel
    ) -> None:
        loss_model._asmnt.asset = asset_model_a
        loss_model.decision_variables = ('my_RV',)
        loss_map = pd.DataFrame(
            {
                'Repair': ['consequence.A', 'consequence.B', 'consequence.F'],
            },
            index=['cmp.A', 'cmp.B', 'cmp.F'],
        )
        loss_model.add_loss_map(loss_map)
        # consequence.A will be for the DS model
        # consequence.B will be for the LF model
        # consequence.C will have no loss parameters defined for it
        # consequence.D should be removed from the DS parameters
        # consequence.E should be removed from the LF parameters
        # consequence.F should be missing
        ds_loss_parameters = pd.DataFrame(
            {
                ('Quantity', 'Unit'): ['1 EA'] * 2,
                ('DV', 'Unit'): ['1 EA'] * 2,
                ('DS1', 'Theta_0'): ['0.00,1.00|0.00,1.00'] * 2,
            },
            index=pd.MultiIndex.from_tuples(
                [('consequence.A', 'my_RV'), ('consequence.D', 'my_RV')]
            ),
        )
        lf_loss_parameters = pd.DataFrame(
            {
                ('Quantity', 'Unit'): ['1 EA'] * 2,
                ('DV', 'Unit'): ['1 EA'] * 2,
                ('Demand', 'Unit'): ['1 EA'] * 2,
                ('LossFunction', 'Theta_0'): ['0.00,1.00|0.00,1.00'] * 2,
            },
            index=pd.MultiIndex.from_tuples(
                [('consequence.B', 'my_RV'), ('consequence.E', 'my_RV')]
            ),
        )
        with pytest.warns(PelicunWarning) as record:
            loss_model.load_model_parameters(
                [ds_loss_parameters, lf_loss_parameters]
            )

        # assert len(record) == 1
        # TODO(JVM): re-enable the line above once we address other
        # warnings, and change indexing to [0] below.

        assert (
            'The loss model does not provide loss information '
            'for the following component(s) in the asset '
            "model: [('consequence.F', 'my_RV')]."
        ) in str(record[-1].message)

    def test__loss_models(self, loss_model: LossModel) -> None:
        models = loss_model._loss_models
        assert len(models) == 2
        assert isinstance(models[0], RepairModel_DS)
        assert isinstance(models[1], RepairModel_LF)

    def test__loss_map(self, loss_model: LossModel) -> None:
        loss_map = pd.DataFrame(
            {
                'Repair': ['consequence_A', 'consequence_B'],
            },
            index=['cmp_A', 'cmp_B'],
        )
        # test setter
        loss_model._loss_map = loss_map
        # test getter
        pd.testing.assert_frame_equal(ensure_value(loss_model._loss_map), loss_map)
        for contained_model in loss_model._loss_models:
            pd.testing.assert_frame_equal(
                ensure_value(contained_model.loss_map), loss_map
            )

    def test__missing(self, loss_model: LossModel) -> None:
        missing = {
            ('missing.component', 'Time'),
            ('missing.component', 'Energy'),
        }
        # test setter
        loss_model._missing = missing
        # test getter
        assert loss_model._missing == missing
        for contained_model in loss_model._loss_models:
            assert contained_model.missing == missing

    def test__ensure_loss_parameter_availability(
        self, assessment_instance: Assessment
    ) -> None:
        loss_model = LossModel(assessment_instance)

        # Only consider `DecisionVariableXYZ`
        loss_model.decision_variables = ('DecisionVariableXYZ',)

        # A, B should be in the ds model
        # C, D should be in the lf model
        # E should be missing

        loss_map = pd.DataFrame(
            {
                'Repair': [f'consequence_{x}' for x in ('A', 'B', 'C', 'D', 'E')],
            },
            index=[f'cmp_{x}' for x in ('A', 'B', 'C', 'D', 'E')],
        )

        loss_model._loss_map = loss_map

        loss_model.ds_model.loss_params = pd.DataFrame(
            index=pd.MultiIndex.from_tuples(
                [
                    ('consequence_A', 'DecisionVariableXYZ'),
                    ('consequence_B', 'DecisionVariableXYZ'),
                ]
            )
        )
        loss_model.lf_model.loss_params = pd.DataFrame(
            index=pd.MultiIndex.from_tuples(
                [
                    ('consequence_C', 'DecisionVariableXYZ'),
                    ('consequence_D', 'DecisionVariableXYZ'),
                ]
            )
        )

        with pytest.warns(PelicunWarning) as record:
            loss_model._ensure_loss_parameter_availability()
            missing = loss_model._missing
        assert missing == {('consequence_E', 'DecisionVariableXYZ')}
        assert len(record) == 1
        assert (
            'The loss model does not provide loss information '
            'for the following component(s) in the asset model: '
            "[('consequence_E', 'DecisionVariableXYZ')]"
        ) in str(record[0].message)

    def test_aggregate_losses_when_no_loss(
        self, assessment_instance: Assessment
    ) -> None:
        # tests that aggregate losses works when there is no loss.
        loss_model = LossModel(assessment_instance)
        loss_model.decision_variables = ('Cost', 'Time', 'Carbon', 'Energy')
        df_agg = loss_model.aggregate_losses()
        assert isinstance(df_agg, pd.DataFrame)
        pd.testing.assert_frame_equal(
            df_agg,
            pd.DataFrame(
                {
                    'repair_cost': 0.00,
                    'repair_carbon': 0.0,
                    'repair_energy': 0.00,
                    'repair_time-sequential': 0.00,
                    'repair_time-parallel': 0.00,
                },
                index=[0],
            ),
        )

    def test__apply_consequence_scaling(
        self, loss_model_with_ones: LossModel
    ) -> None:
        # When only `dv` is provided
        scaling_conditions = {'dv': 'Cost'}
        scaling_factor = 2.00

        loss_model_with_ones._apply_consequence_scaling(
            scaling_conditions, scaling_factor
        )

        for loss_model in loss_model_with_ones._loss_models:
            assert loss_model.sample is not None
            mask = loss_model.sample.columns.get_level_values('dv') == 'Cost'
            assert np.all(loss_model.sample.iloc[:, mask] == 2.00)
            assert np.all(loss_model.sample.iloc[:, ~mask] == 1.00)
            loss_model.sample.iloc[:, :] = 1.00

        scaling_conditions = {'dv': 'Carbon', 'loc': '1', 'uid': 'uid2'}
        scaling_factor = 2.00
        loss_model_with_ones._apply_consequence_scaling(
            scaling_conditions, scaling_factor
        )

        for loss_model in loss_model_with_ones._loss_models:
            assert loss_model.sample is not None
            mask = np.full(len(loss_model.sample.columns), fill_value=True)
            mask &= loss_model.sample.columns.get_level_values('dv') == 'Carbon'
            mask &= loss_model.sample.columns.get_level_values('loc') == '1'
            mask &= loss_model.sample.columns.get_level_values('uid') == 'uid2'
            assert np.all(loss_model.sample.iloc[:, mask] == 2.00)
            assert np.all(loss_model.sample.iloc[:, ~mask] == 1.00)

    def test_aggregate_losses_combination(
        self, assessment_instance: Assessment
    ) -> None:
        # The test sets up a very simple loss calculation from
        # scratch, only defining essential parameters.

        # demand
        sample_size = 5
        demand_marginal_parameters = pd.DataFrame(
            {
                ('PIH', '0', '1'): ['in', 7.00],
                ('PWS', '0', '1'): ['mph', 50.0],
            },
            index=['Units', 'Theta_0'],
        ).T
        perfect_corr = pd.DataFrame(
            np.ones((2, 2)),
            columns=demand_marginal_parameters.index,
            index=demand_marginal_parameters.index,
        )
        assessment_instance.demand.load_model(
            {'marginals': demand_marginal_parameters, 'correlation': perfect_corr}
        )
        assessment_instance.demand.generate_sample({'SampleSize': sample_size})

        # asset
        assessment_instance.asset.cmp_marginal_params = pd.DataFrame(
            {
                'Theta_0': (1.0, 1.0),
            },
            index=pd.MultiIndex.from_tuples(
                (('wind.comp', '0', '1', '0'), ('flood.comp', '0', '1', '0')),
                names=('cmp', 'loc', 'dir', 'uid'),
            ),
        )
        assessment_instance.asset.generate_cmp_sample()

        # no damage estimation needed since we only use loss functions

        # loss

        assessment_instance.loss.decision_variables = ('Cost',)
        assessment_instance.loss.add_loss_map(loss_map_policy='fill')
        assessment_instance.loss.load_model_parameters(
            [
                (
                    'pelicun/tests/basic/data/model/'
                    'test_LossModel/loss_function_wind.csv'
                ),
                (
                    'pelicun/tests/basic/data/model/'
                    'test_LossModel/loss_function_flood.csv'
                ),
            ]
        )

        assessment_instance.loss.calculate()

        # individual losses
        l1, l2 = ensure_value(assessment_instance.loss.lf_model.sample).iloc[0, :]
        # combined loss, result of interpolation
        l_comb = 0.904

        file_path = file_io.substitute_default_path(
            ['PelicunDefault/Hazus Hurricane Wind/combine_wind_flood.csv']
        )[0]
        assert isinstance(file_path, str)
        combination_array = pd.read_csv(
            file_path,
            index_col=None,
            header=None,
        ).to_numpy()
        loss_combination = {
            'Cost': {
                ('wind.comp', 'flood.comp'): combination_array,
            },
        }

        agg_df, _ = assessment_instance.loss.aggregate_losses(
            loss_combination=loss_combination, future=True
        )
        assert isinstance(agg_df, pd.DataFrame)
        pd.testing.assert_frame_equal(
            agg_df, pd.DataFrame([l_comb] * 5, columns=['repair_cost'])
        )

        # verify interpolation with some manual checks
        lower, higher = combination_array[8:10, 4]
        assert lower <= l_comb <= higher
        assert l2 == combination_array[0, 4]
        assert combination_array[8, 0] <= l1 <= combination_array[9, 0]

    def test_aggregate_losses_thresholds(
        self, loss_model_with_ones: LossModel
    ) -> None:
        # Row 0 has the value of 1.0 in all columns.
        # Adjust rows 1 and 2 to have the values 2.0 and 3.0, for
        # testing.
        assert loss_model_with_ones.ds_model.sample is not None
        assert loss_model_with_ones.lf_model.sample is not None
        loss_model_with_ones.decision_variables = ('Cost', 'Carbon')
        loss_model_with_ones.dv_units = {'Cost': 'USD_2011', 'Carbon': 'kg'}
        loss_model_with_ones.ds_model.sample.iloc[1, :] = 2.00
        loss_model_with_ones.ds_model.sample.iloc[2, :] = 3.00
        loss_model_with_ones.lf_model.sample.iloc[1, :] = 2.00
        loss_model_with_ones.lf_model.sample.iloc[2, :] = 3.00
        # Instantiate a RandomVariableRegistry to pass as an argument
        # to the method.
        rv_reg = uq.RandomVariableRegistry(loss_model_with_ones._asmnt.options.rng)
        # Add a threshold for `Cost`
        rv_reg.add_RV(
            uq.rv_class_map('deterministic')(name='Cost', theta=np.array((400.00,)))  # type: ignore
        )
        # Add a threshold for `Carbon`
        rv_reg.add_RV(
            uq.rv_class_map('deterministic')(
                name='Carbon',
                theta=np.array((100.00,)),  # type: ignore
            )
        )
        df_agg, exceedance_bool_df = loss_model_with_ones.aggregate_losses(
            replacement_configuration=(rv_reg, {'Cost': 0.50, 'Carbon': 1.00}),
            future=True,
        )
        assert isinstance(df_agg, pd.DataFrame)
        assert isinstance(exceedance_bool_df, pd.DataFrame)
        df_agg_expected = pd.DataFrame(
            {
                'repair_carbon': [96.00, 100.00, 100.00],
                'repair_cost': [96.00, 400.00, 400.00],
            }
        )
        exceedance_bool_df_expected = pd.DataFrame(
            {'Cost': [False, False, True], 'Carbon': [False, True, True]}
        )
        pd.testing.assert_frame_equal(df_agg, df_agg_expected)
        pd.testing.assert_frame_equal(
            exceedance_bool_df, exceedance_bool_df_expected
        )

    def test_consequence_scaling(self, loss_model_with_ones: LossModel) -> None:
        loss_model_with_ones.consequence_scaling(
            'pelicun/tests/basic/data/model/test_LossModel/scaling_specification.csv'
        )

        expected_ds = (
            pd.read_csv(
                'pelicun/tests/basic/data/model/test_LossModel/scaled_losses_ds.csv',
                dtype={
                    'dv': str,
                    'loss': str,
                    'dmg': str,
                    'ds': str,
                    'loc': str,
                    'dir': str,
                    'uid': str,
                },
            )
            .set_index(['dv', 'loss', 'dmg', 'ds', 'loc', 'dir', 'uid'])
            .T.astype(float)
        )
        expected_ds.index = pd.RangeIndex(range(len(expected_ds)))  # type: ignore
        pd.testing.assert_frame_equal(
            loss_model_with_ones.ds_model.sample,  # type: ignore
            expected_ds,
        )

        expected_lf = (
            pd.read_csv(
                'pelicun/tests/basic/data/model/test_LossModel/scaled_losses_lf.csv',
                dtype={
                    'dv': str,
                    'loss': str,
                    'dmg': str,
                    'loc': str,
                    'dir': str,
                    'uid': str,
                },
            )
            .set_index(['dv', 'loss', 'dmg', 'loc', 'dir', 'uid'])
            .T.astype(float)
        )
        expected_lf.index = pd.RangeIndex(range(len(expected_lf)))  # type: ignore
        pd.testing.assert_frame_equal(
            loss_model_with_ones.lf_model.sample,  # type: ignore
            expected_lf,
        )


class TestRepairModel_Base(TestPelicunModel):
    def test___init__(self, assessment_instance: Assessment) -> None:
        repair_model = RepairModel_DS(assessment_instance)
        with pytest.raises(AttributeError):
            repair_model.xyz = 123  # type: ignore

    def test_drop_unused_loss_parameters(
        self, assessment_instance: Assessment
    ) -> None:
        base_model = RepairModel_DS(assessment_instance)
        loss_map = pd.DataFrame(
            {
                'Repair': ['consequence_A', 'consequence_B'],
            },
            index=['cmp_A', 'cmp_B'],
        )
        # without loss_params, it should do nothing
        base_model.drop_unused_loss_parameters(loss_map)
        base_model.loss_params = pd.DataFrame(
            index=[f'consequence_{x}' for x in ('A', 'B', 'C', 'D')]
        )
        base_model.drop_unused_loss_parameters(loss_map)
        pd.testing.assert_frame_equal(
            base_model.loss_params,
            pd.DataFrame(index=[f'consequence_{x}' for x in ('A', 'B')]),
        )

    def test__remove_incomplete_components(
        self, assessment_instance: Assessment
    ) -> None:
        base_model = RepairModel_DS(assessment_instance)
        # without loss_params, it should do nothing
        base_model.remove_incomplete_components()
        # without incomplete, it should do nothing
        loss_params = pd.DataFrame(
            index=[f'consequence_{x}' for x in ('A', 'B', 'C', 'D')]
        )
        base_model.loss_params = loss_params
        base_model.remove_incomplete_components()
        pd.testing.assert_frame_equal(
            base_model.loss_params,
            loss_params,
        )
        base_model.loss_params = pd.DataFrame(
            {('Incomplete', ''): [0, 0, 0, 1]},
            index=[f'consequence_{x}' for x in ('A', 'B', 'C', 'D')],
        )
        # Now entry D should be gone
        base_model.remove_incomplete_components()
        pd.testing.assert_frame_equal(
            base_model.loss_params,
            pd.DataFrame(
                {('Incomplete', ''): [0, 0, 0]},
                index=[f'consequence_{x}' for x in ('A', 'B', 'C')],
            ),
        )

    def test__get_available(self, assessment_instance: Assessment) -> None:
        base_model = RepairModel_DS(assessment_instance)
        base_model.loss_params = pd.DataFrame(index=['cmp.A', 'cmp.B', 'cmp.C'])
        assert base_model.get_available() == {'cmp.A', 'cmp.B', 'cmp.C'}


class TestRepairModel_DS(TestRepairModel_Base):
    def test_convert_loss_parameter_units(
        self, assessment_instance: Assessment
    ) -> None:
        ds_model = RepairModel_DS(assessment_instance)
        ds_model.loss_params = pd.DataFrame(
            {
                ('Quantity', 'Unit'): ['1 test_two', '1 EA'],
                ('DV', 'Unit'): ['test_three', 'test_three'],
                ('DS1', 'Theta_0'): ['200.00,100.00|10.00,20.00', '100.00'],
                ('DS1', 'Theta_1'): [0.20, None],
                ('DS1', 'Family'): ['lognormal', None],
            },
            index=pd.MultiIndex.from_tuples([('cmpA', 'Cost'), ('cmpB', 'Cost')]),
        )

        ds_model.convert_loss_parameter_units()

        # DVs are scaled by 3/2, quantities by 2
        pd.testing.assert_frame_equal(
            ds_model.loss_params,
            pd.DataFrame(
                {
                    ('Quantity', 'Unit'): ['1 test_two', '1 EA'],
                    ('DV', 'Unit'): ['test_three', 'test_three'],
                    ('DS1', 'Theta_0'): ['300,150|20,40', 300.0],
                    ('DS1', 'Theta_1'): [0.20, None],
                    ('DS1', 'Family'): ['lognormal', None],
                },
                index=pd.MultiIndex.from_tuples(
                    [('cmpA', 'Cost'), ('cmpB', 'Cost')]
                ),
            ),
        )

    def test__drop_unused_damage_states(
        self, assessment_instance: Assessment
    ) -> None:
        ds_model = RepairModel_DS(assessment_instance)
        loss_params = pd.DataFrame(
            {
                ('DS1', 'Theta_0'): [1.0, 1.0, 1.0, 1.0],
                ('DS2', 'Theta_0'): [1.0, 1.0, 1.0, None],
                ('DS3', 'Theta_0'): [1.0, 1.0, None, None],
                ('DS4', 'Theta_0'): [1.0, None, None, None],
                ('DS5', 'Theta_0'): [None, None, None, None],
                ('DS6', 'Theta_0'): [None, None, None, None],
                ('DS7', 'Theta_0'): [None, None, None, None],
            }
        )
        ds_model.loss_params = loss_params
        ds_model.drop_unused_damage_states()
        pd.testing.assert_frame_equal(ds_model.loss_params, loss_params.iloc[:, 0:4])

    def test__create_DV_RVs(self, assessment_instance: Assessment) -> None:
        assessment_instance.options.rho_cost_time = 0.30
        ds_model = RepairModel_DS(assessment_instance)
        ds_model.decision_variables = ('Cost', 'Time')
        ds_model.missing = {('cmp.B', 'Cost'), ('cmp.B', 'Time')}
        ds_model.loss_map = pd.DataFrame(
            {
                'Repair': ['cmp.A', 'cmp.B', 'cmp.C', 'cmp.D', 'cmp.E'],
            },
            index=['cmp.A', 'cmp.B', 'cmp.C', 'cmp.D', 'cmp.E'],
        )
        # cmp.B is marked as missing, cmp.C is intended for the LF
        # model.
        # cmp.D has `|` in Theta_0 which should be treated as 1.00
        # cmp.E has deterministic loss.
        ds_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA', '1 EA', '1 EA', '1 EA'],
                ('Quantity', 'Unit'): ['1 EA', '1 EA', '1 EA', '1 EA'],
                ('DS1', 'Family'): ['normal', 'normal', 'normal', None],
                ('DS1', 'Theta_0'): [1.00, 1.00, '4.0,2.0|5.0,1.0', 1.00],
                ('DS1', 'Theta_1'): [1.00, 1.00, 1.00, None],
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ('cmp.A', 'Cost'),
                    ('cmp.A', 'Time'),
                    ('cmp.D', 'Cost'),
                    ('cmp.E', 'Cost'),
                ]
            ),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])

        cases = pd.MultiIndex.from_tuples(
            [
                ('cmp.A', '0', '1', '0', '1'),
                ('cmp.B', '0', '1', '0', '1'),  # marked as missing
                ('cmp.C', '0', '1', '0', '1'),  # no loss parameters
                ('cmp.D', '0', '1', '0', '1'),  # `|` in Theta_0
                ('cmp.E', '0', '1', '0', '1'),  # Deterministic loss
            ],
            names=['cmp', 'loc', 'dir', 'uid', 'ds'],
        )
        rv_reg = ds_model._create_DV_RVs(cases)
        assert rv_reg is not None
        for key in (
            'Cost-cmp.A-1-0-1-0',
            'Time-cmp.A-1-0-1-0',
            'Cost-cmp.D-1-0-1-0',
        ):
            assert key in rv_reg.RV
        assert len(rv_reg.RV) == 3
        assert isinstance(rv_reg.RV['Cost-cmp.A-1-0-1-0'], uq.NormalRandomVariable)
        assert isinstance(rv_reg.RV['Time-cmp.A-1-0-1-0'], uq.NormalRandomVariable)
        assert isinstance(rv_reg.RV['Cost-cmp.D-1-0-1-0'], uq.NormalRandomVariable)
        assert np.all(
            rv_reg.RV['Cost-cmp.A-1-0-1-0'].theta[0:2] == np.array((1.0, 1.0))  # type: ignore
        )
        assert np.all(
            rv_reg.RV['Time-cmp.A-1-0-1-0'].theta[0:2] == np.array((1.0, 1.0))  # type: ignore
        )
        assert np.all(
            rv_reg.RV['Cost-cmp.D-1-0-1-0'].theta[0:2] == np.array([1.0, 1.0])  # type: ignore
        )
        assert 'DV-cmp.A-1-0-1-0_set' in rv_reg.RV_set
        np.all(
            rv_reg.RV_set['DV-cmp.A-1-0-1-0_set'].Rho()
            == np.array(((1.0, 0.3), (0.3, 1.0)))
        )
        assert len(rv_reg.RV_set) == 1

    def test__create_DV_RVs_all_deterministic(
        self, assessment_instance: Assessment
    ) -> None:
        ds_model = RepairModel_DS(assessment_instance)
        ds_model.decision_variables = ('myRV',)
        ds_model.missing = set()
        ds_model.loss_map = pd.DataFrame(
            {'Repair': ['cmp.A']},
            index=['cmp.A'],
        )
        ds_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA'],
                ('Quantity', 'Unit'): ['1 EA'],
                ('DS1', 'Family'): [None],
                ('DS1', 'Theta_0'): [1.00],
            },
            index=pd.MultiIndex.from_tuples([('cmp.A', 'myRV')]),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])

        cases = pd.MultiIndex.from_tuples(
            [('cmp.A', '0', '1', '0', '1')],
            names=['cmp', 'loc', 'dir', 'uid', 'ds'],
        )
        rv_reg = ds_model._create_DV_RVs(cases)

        assert rv_reg is None

    def test__calc_median_consequence_no_locs(
        self, assessment_instance: Assessment
    ) -> None:
        # Test the method when the eco_qnt dataframe's columns do not
        # contain `loc` information.

        ds_model = RepairModel_DS(assessment_instance)
        eco_qnt = pd.DataFrame(
            {
                ('cmp.A', '0'): [0.00, 0.00, 1.00],
                ('cmp.A', '1'): [1.00, 0.00, 0.00],
                ('cmp.A', '2'): [0.00, 1.00, 0.00],
                ('cmp.B', '1'): [0.00, 1.00, 0.00],
                ('cmp.B', '2'): [1.00, 0.00, 0.00],
            }
        ).rename_axis(columns=['cmp', 'ds'])
        ds_model.decision_variables = ('my_DV',)
        # cmp.A should be available and we should get medians.
        # missing_cmp will be marked as missing
        # is_for_LF_model represents a component->consequence pair
        # that is intended for processing by the loss function model
        # and should be ignored by the damage state model.
        ds_model.loss_map = pd.DataFrame(
            {
                'Repair': ['cmp.A', 'cmp.B', 'missing_cmp', 'is_for_LF_model'],
            },
            index=['cmp.A', 'cmp.B', 'missing_consequence', 'LF_consequence'],
        )

        # DS3 is in the loss parameters but has not been triggered.
        ds_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA', '1 EA'],
                ('Quantity', 'Unit'): ['1 EA', '1 EA'],
                ('DS1', 'Family'): [None, 'normal'],
                ('DS1', 'Theta_0'): [100.00, 12345.00],
                ('DS1', 'Theta_1'): [None, 0.30],
                ('DS2', 'Family'): [None, 'normal'],
                ('DS2', 'Theta_0'): [200.00, '2.00,1.00|5.00,10.00'],
                ('DS2', 'Theta_1'): [None, 0.30],
                ('DS3', 'Family'): [None, 'normal'],
                ('DS3', 'Theta_0'): [200.00, '2.00,1.00|5.00,10.00'],
                ('DS3', 'Theta_1'): [None, 0.30],
            },
            index=pd.MultiIndex.from_tuples(
                [('cmp.A', 'my_DV'), ('cmp.B', 'my_DV')]
            ),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])
        ds_model.missing = {('missing_cmp', 'my_DV')}
        medians = ds_model._calc_median_consequence(eco_qnt)
        assert len(medians) == 1
        assert 'my_DV' in medians
        pd.testing.assert_frame_equal(
            medians['my_DV'],
            pd.DataFrame(
                {
                    ('cmp.A', '1'): [100.00, 100.00, 100.00],
                    ('cmp.A', '2'): [200.00, 200.00, 200.00],
                    ('cmp.B', '1'): [1.00, 1.00, 1.00],
                    ('cmp.B', '2'): [2.00, 2.00, 2.00],
                }
            ).rename_axis(columns=['cmp', 'ds']),
        )

        #
        # edge cases
        #

        # random variable not supported
        ds_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA'],
                ('Quantity', 'Unit'): ['1 EA'],
                ('DS1', 'Family'): ['multilinear_CDF'],
                ('DS1', 'Theta_0'): ['0.00,1.00|0.00,1.00'],
                ('DS1', 'Theta_1'): [0.30],
            },
            index=pd.MultiIndex.from_tuples([('cmp.A', 'my_DV')]),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])
        with pytest.raises(
            ValueError,
            match='Loss Distribution of type multilinear_CDF not supported.',
        ):
            ds_model._calc_median_consequence(eco_qnt)

    def test__calc_median_consequence_locs(
        self, assessment_instance: Assessment
    ) -> None:
        # Test the method when the eco_qnt dataframe's columns contain
        # `loc` information.

        ds_model = RepairModel_DS(assessment_instance)
        eco_qnt = pd.DataFrame(
            {
                ('cmp.A', '0', '1'): [0.00, 0.00, 1.00],
                ('cmp.A', '1', '1'): [1.00, 0.00, 0.00],
            }
        ).rename_axis(columns=['cmp', 'ds', 'loc'])
        ds_model.decision_variables = ('my_DV',)
        # cmp.A should be available and we should get medians.
        # missing_cmp will be marked as missing
        # is_for_LF_model represents a component->consequence pair
        # that is intended for processing by the loss function model
        # and should be ignored by the damage state model.
        ds_model.loss_map = pd.DataFrame(
            {
                'Repair': ['cmp.A'],
            },
            index=['cmp.A'],
        )

        # DS3 is in the loss parameters but has not been triggered.
        ds_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA'],
                ('Quantity', 'Unit'): ['1 EA'],
                ('DS1', 'Family'): [None],
                ('DS1', 'Theta_0'): [100.00],
                ('DS1', 'Theta_1'): [None],
                ('DS2', 'Family'): [None],
                ('DS2', 'Theta_0'): [200.00],
                ('DS2', 'Theta_1'): [None],
                ('DS3', 'Family'): [None],
                ('DS3', 'Theta_0'): [200.00],
                ('DS3', 'Theta_1'): [None],
            },
            index=pd.MultiIndex.from_tuples([('cmp.A', 'my_DV')]),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])
        ds_model.missing = set()
        medians = ds_model._calc_median_consequence(eco_qnt)
        assert len(medians) == 1
        assert 'my_DV' in medians
        pd.testing.assert_frame_equal(
            medians['my_DV'],
            pd.DataFrame(
                {
                    ('cmp.A', '1', '1'): [100.00, 100.00, 100.00],
                }
            ).rename_axis(columns=['cmp', 'ds', 'loc']),
        )


class TestRepairModel_LF(TestRepairModel_Base):
    def test_convert_loss_parameter_units(
        self, assessment_instance: Assessment
    ) -> None:
        lf_model = RepairModel_LF(assessment_instance)
        lf_model.loss_params = pd.DataFrame(
            {
                ('Demand', 'Unit'): ['inps2', 'g'],
                ('DV', 'Unit'): ['test_three', 'test_three'],
                ('LossFunction', 'Theta_0'): [
                    '1.00,1.00|1.00,1.00',
                    '1.00,1.00|1.00,1.00',
                ],
            },
            index=pd.MultiIndex.from_tuples([('cmpA', 'Cost'), ('cmpB', 'Cost')]),
        )

        lf_model.convert_loss_parameter_units()

        pd.testing.assert_frame_equal(
            lf_model.loss_params,
            pd.DataFrame(
                {
                    ('Demand', 'Unit'): ['inps2', 'g'],
                    ('DV', 'Unit'): ['test_three', 'test_three'],
                    ('LossFunction', 'Theta_0'): [
                        '3,3|0.0254,0.0254',
                        '3,3|9.80665,9.80665',
                    ],
                },
                index=pd.MultiIndex.from_tuples(
                    [('cmpA', 'Cost'), ('cmpB', 'Cost')]
                ),
            ),
        )

    def test__calc_median_consequence(self, assessment_instance: Assessment) -> None:
        lf_model = RepairModel_LF(assessment_instance)

        performance_group = pd.DataFrame(
            {
                'Blocks': [1],
            },
            index=pd.MultiIndex.from_tuples([(('cmp.A', 'dv.A'), '0', '1', '0')]),
        )
        loss_map = {'cmp.A': 'cmp.A'}
        required_edps = {(('cmp.A', 'dv.A'), '0', '1', '0'): 'PFA-1-1'}
        demand_dict = {'PFA-1-1': np.array((1.00, 2.00, 3.00))}
        cmp_sample = {
            ('cmp.A', '0', '1', '0'): pd.Series(
                np.array((10.00, 20.00, 30.00)), name=('cmp.A', '0', '1', '0')
            )
        }
        lf_model.loss_params = pd.DataFrame(
            {
                ('LossFunction', 'Theta_0'): ['0.00,1.00|0.00,10.00'],
            },
            index=pd.MultiIndex.from_tuples(
                [('cmp.A', 'dv.A')], names=['Loss Driver', 'Decsision Variable']
            ),
        )
        medians = lf_model._calc_median_consequence(
            performance_group, loss_map, required_edps, demand_dict, cmp_sample
        )
        pd.testing.assert_frame_equal(
            medians,
            pd.DataFrame(
                {('dv.A', 'cmp.A', 'cmp.A', '0', '1', '0', '0'): [1.0, 4.0, 9.0]},
            ).rename_axis(
                columns=['dv', 'loss', 'dmg', 'loc', 'dir', 'uid', 'block']
            ),
        )
        # test small interpolation domain warning
        demand_dict = {'PFA-1-1': np.array((1.00, 2.00, 1e3))}
        with pytest.raises(
            ValueError,
            match=re.escape(
                'Loss function interpolation for consequence '
                '`cmp.A-dv.A` has failed. Ensure a sufficient '
                'interpolation domain  for the X values '
                '(those after the `|` symbol)  and verify '
                'the X-value and Y-value lengths match.'
            ),
        ):
            lf_model._calc_median_consequence(
                performance_group, loss_map, required_edps, demand_dict, cmp_sample
            )

    def test__create_DV_RVs(self, assessment_instance: Assessment) -> None:
        assessment_instance.options.rho_cost_time = 0.50
        lf_model = RepairModel_LF(assessment_instance)
        lf_model.decision_variables = ('Cost', 'Time')
        lf_model.missing = set()
        lf_model.loss_map = pd.DataFrame(
            {
                'Repair': ['cmp.A', 'cmp.B'],
            },
            index=['cmp.A', 'cmp.B'],
        )
        lf_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA', '1 EA', '1 EA'],
                ('Quantity', 'Unit'): ['1 EA', '1 EA', '1 EA'],
                ('LossFunction', 'Family'): ['normal', 'normal', None],
                ('LossFunction', 'Theta_0'): [
                    '0.0,1.0|0.0,1.0',
                    '0.0,1.0|0.0,1.0',
                    '0.0,1.0|0.0,1.0',
                ],
                ('LossFunction', 'Theta_1'): [0.3, 0.3, None],
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ('cmp.A', 'Cost'),
                    ('cmp.A', 'Time'),
                    ('cmp.B', 'Cost'),
                ]
            ),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])

        cases = pd.MultiIndex.from_tuples(
            [
                ('Cost', 'cmp.A', 'cmp.A', '0', '1', '0', '1'),
                ('Time', 'cmp.A', 'cmp.A', '0', '1', '0', '1'),
                ('Cost', 'cmp.B', 'cmp.B', '0', '1', '0', '1'),
            ],
            names=['dv', 'loss', 'dmg', 'loc', 'dir', 'uid', 'block'],
        )
        rv_reg = lf_model._create_DV_RVs(cases)
        assert rv_reg is not None
        for key in (
            'Cost-cmp.A-cmp.A-0-1-0-1',
            'Time-cmp.A-cmp.A-0-1-0-1',
        ):
            assert key in rv_reg.RV
        assert len(rv_reg.RV) == 2
        assert isinstance(
            rv_reg.RV['Cost-cmp.A-cmp.A-0-1-0-1'], uq.NormalRandomVariable
        )
        assert isinstance(
            rv_reg.RV['Time-cmp.A-cmp.A-0-1-0-1'], uq.NormalRandomVariable
        )
        assert np.all(
            rv_reg.RV['Cost-cmp.A-cmp.A-0-1-0-1'].theta[0:2] == np.array((1.0, 0.3))  # type: ignore
        )
        assert np.all(
            rv_reg.RV['Time-cmp.A-cmp.A-0-1-0-1'].theta[0:2] == np.array((1.0, 0.3))  # type: ignore
        )
        assert 'DV-cmp.A-cmp.A-0-1-0-1_set' in rv_reg.RV_set
        np.all(
            rv_reg.RV_set['DV-cmp.A-cmp.A-0-1-0-1_set'].Rho()
            == np.array(((1.0, 0.5), (0.5, 1.0)))
        )
        assert len(rv_reg.RV_set) == 1

    def test__create_DV_RVs_no_rv_case(
        self, assessment_instance: Assessment
    ) -> None:
        # Special case where there is no need for RVs

        lf_model = RepairModel_LF(assessment_instance)
        lf_model.decision_variables = ('Cost', 'Time')
        lf_model.missing = set()
        lf_model.loss_map = pd.DataFrame(
            {
                'Repair': ['cmp.B'],
            },
            index=['cmp.B'],
        )
        lf_model.loss_params = pd.DataFrame(
            {
                ('DV', 'Unit'): ['1 EA'],
                ('Quantity', 'Unit'): ['1 EA'],
                ('LossFunction', 'Family'): [None],
                ('LossFunction', 'Theta_0'): ['0.0,1.0|0.0,1.0'],
                ('LossFunction', 'Theta_1'): [None],
            },
            index=pd.MultiIndex.from_tuples(
                [
                    ('cmp.B', 'Cost'),
                ]
            ),
        ).rename_axis(index=['Loss Driver', 'Decision Variable'])

        cases = pd.MultiIndex.from_tuples(
            [
                ('Cost', 'cmp.B', 'cmp.B', '0', '1', '0', '1'),
            ],
            names=['dv', 'loss', 'dmg', 'loc', 'dir', 'uid', 'block'],
        )
        rv_reg = lf_model._create_DV_RVs(cases)
        assert rv_reg is None


def test__prep_constant_median_DV() -> None:
    median = 10.00
    constant_median_dv = model.loss_model._prep_constant_median_DV(median)
    assert constant_median_dv() == median
    values = (1.0, 2.0, 3.0, 4.0, 5.0)
    for value in values:
        assert constant_median_dv(value) == 10.00


def test__prep_bounded_multilinear_median_DV() -> None:
    medians = np.array((1.00, 2.00, 3.00, 4.00, 5.00))
    quantities = np.array((0.00, 1.00, 2.00, 3.00, 4.00))
    f = model.loss_model._prep_bounded_multilinear_median_DV(medians, quantities)

    result = f(2.5)
    expected = 3.5
    assert result == expected

    result = f(0.00)
    expected = 1.00
    assert result == expected

    result = f(4.00)
    expected = 5.0
    assert result == expected

    result = f(-1.00)
    expected = 1.00
    assert result == expected

    result = f(5.00)
    expected = 5.00
    assert result == expected

    result_list = f([2.5, 3.5])
    expected_list = [3.5, 4.5]
    assert np.allclose(result_list, expected_list)

    with pytest.raises(
        ValueError,
        match=(
            'A bounded linear median Decision Variable function '
            'called without specifying the quantity '
            'of damaged components'
        ),
    ):
        f(None)


def test__is_for_lf_model() -> None:
    positive_case = pd.DataFrame(
        {
            ('LossFunction', 'Theta_0'): [0.5],
        },
        index=pd.Index(['cmp.1'], name='ID'),
    )

    negative_case = pd.DataFrame(
        {
            ('DS1', 'Theta_0'): [0.50],
        },
        index=pd.Index(['cmp.1'], name='ID'),
    )

    assert _is_for_lf_model(positive_case) is True
    assert _is_for_lf_model(negative_case) is False


def test__is_for_ds_model() -> None:
    positive_case = pd.DataFrame(
        {
            ('DS1', 'Theta_0'): [0.50],
        },
        index=pd.Index(['cmp.1'], name='ID'),
    )

    negative_case = pd.DataFrame(
        {
            ('LossFunction', 'Theta_0'): [0.5],
        },
        index=pd.Index(['cmp.1'], name='ID'),
    )

    assert _is_for_ds_model(positive_case) is True
    assert _is_for_ds_model(negative_case) is False



# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay

"""Provides a command-line interface for Pelicun."""

from __future__ import annotations

import argparse
import logging
from datetime import datetime

from pelicun.tools.dlml import dlml_update
from pelicun.tools.regional_sim import regional_sim


def setup_dlml_logging(log_file: str | bool | None = None) -> None:  # noqa: FBT001
    """
    Configure logging for DLML operations.

    Parameters
    ----------
    log_file : str, optional
        Path to log file. If True, creates timestamped file. If None, no file logging.
    """
    logger = logging.getLogger('pelicun.dlml')

    # Only add handlers if none exist (avoid duplicates)
    if not logger.handlers:
        # Always add stdout handler for CLI operations
        stdout_handler = logging.StreamHandler()
        stdout_formatter = logging.Formatter('%(message)s')
        stdout_handler.setFormatter(stdout_formatter)
        logger.addHandler(stdout_handler)

        # Add file handler if requested
        if log_file:
            if log_file is True:  # --log without filename
                timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')  # noqa: DTZ005
                log_file = f'dlml_update_{timestamp}.log'

            file_handler = logging.FileHandler(log_file)
            file_formatter = logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            )
            file_handler.setFormatter(file_formatter)
            logger.addHandler(file_handler)

            print(f'Logging to file: {log_file}')  # noqa: T201

        logger.setLevel(logging.INFO)


def main() -> None:
    """
    Provide main command-line interface for Pelicun.

    This function dispatches subcommands.

    """
    # Define comprehensive usage examples
    examples = """
Examples:
  Regional Simulation:
    pelicun regional_sim                          # Use default config file (inputRWHALE.json)
    pelicun regional_sim my_config.json           # Use custom config file
    pelicun regional_sim -n 4                     # Use 4 CPU cores with default config
    pelicun regional_sim my_config.json -n 8      # Use custom config with 8 CPU cores

  DLML Data Management:
    pelicun dlml update                           # Update to latest DLML version
    pelicun dlml update latest                    # Same as above (explicit)
    pelicun dlml update v1.2.0                   # Update to specific version
    pelicun dlml update "commit abc1234"          # Update to specific commit SHA
    pelicun dlml update --no-cache latest         # Force re-download without caching
    pelicun dlml update --no-cache v1.2.0         # Update to version without caching
    pelicun dlml update --no-cache "commit def567" # Update to commit without caching

  Getting Help:
    pelicun --help                                # Show this help message
    pelicun regional_sim --help                   # Show regional_sim specific help
    pelicun dlml --help                           # Show dlml specific help
"""

    # Main parser
    parser = argparse.ArgumentParser(
        description='Main command-line interface for Pelicun.',
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers(
        dest='subcommand', required=True, help='Available subcommands'
    )

    # Create the parser for the "regional_sim" subcommand
    parser_regional = subparsers.add_parser(
        'regional_sim', help='Perform a regional-scale disaster impact simulation.'
    )

    # Add the arguments specific to regional_sim
    parser_regional.add_argument(
        'config_file',
        nargs='?',
        default='inputRWHALE.json',
        help='Path to the input configuration JSON file. '
        "Defaults to 'inputRWHALE.json'.",
    )
    parser_regional.add_argument(
        '-n',
        '--num-cores',
        type=int,
        default=None,
        help='Number of CPU cores to use for parallel processing. '
        'Defaults to all available cores minus one.',
    )
    # Associate the regional_sim function with this subparser
    parser_regional.set_defaults(func=regional_sim)

    # Create the parser for the "dlml" subcommand
    parser_dlml = subparsers.add_parser(
        'dlml', help='Update DLML (Damage and Loss Model Library) data files.'
    )

    # Add the arguments specific to dlml
    parser_dlml.add_argument(
        'action',
        choices=['update'],
        help='Action to perform. Currently only "update" is supported.',
    )
    parser_dlml.add_argument(
        'target',
        nargs='?',
        default='latest',
        help='Version tag (e.g., v1.2.0) or "latest" for the latest release. '
        'Use "commit <sha>" to specify a commit SHA.',
    )
    parser_dlml.add_argument(
        '--no-cache',
        action='store_true',
        help='Disable caching to force re-download of all files.',
    )
    parser_dlml.add_argument(
        '--log',
        nargs='?',
        const=True,
        metavar='LOGFILE',
        help='Save detailed log to specified file. If no filename provided, '
        'creates dlml_update_TIMESTAMP.log in current directory.',
    )
    # Associate the dlml_update function with this subparser
    parser_dlml.set_defaults(func=dlml_update)

    # Parse the arguments from the command line
    args = parser.parse_args()

    # Call the function associated with the chosen subcommand
    if args.subcommand == 'regional_sim':
        args.func(config_file=args.config_file, num_cores=args.num_cores)
    elif args.subcommand == 'dlml':
        # Setup logging for DLML operations
        setup_dlml_logging(log_file=args.log)

        # Handle dlml arguments
        use_cache = not args.no_cache
        if args.target.startswith('commit '):
            commit_sha = args.target.split(' ', 1)[1]
            args.func(commit=commit_sha, use_cache=use_cache)
        else:
            args.func(version=args.target, use_cache=use_cache)



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the PelicunModel class."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import pytest

from pelicun import model
from pelicun.tests.basic.test_model import TestModelModule

if TYPE_CHECKING:
    from pelicun.assessment import Assessment
    from pelicun.model.pelicun_model import PelicunModel


class TestPelicunModel(TestModelModule):
    @pytest.fixture
    def pelicun_model(self, assessment_instance: Assessment) -> PelicunModel:
        return deepcopy(model.PelicunModel(assessment_instance))

    def test_init(self, pelicun_model: PelicunModel) -> None:
        assert pelicun_model.log

    def test__convert_marginal_params(self, pelicun_model: PelicunModel) -> None:
        # one row, only Theta_0, no conversion
        marginal_params = pd.DataFrame(
            [['1.0']],
            columns=['Theta_0'],
            index=pd.MultiIndex.from_tuples(
                (('A', '0', '1'),), names=('cmp', 'loc', 'dir')
            ),
        )
        units = pd.Series(['ea'], index=marginal_params.index)
        arg_units = None
        res = pelicun_model._convert_marginal_params(
            marginal_params, units, arg_units
        )

        # >>> res
        #             Theta_0
        # cmp loc dir
        # A   0   1       1.0

        assert 'Theta_0' in res.columns
        assert res.to_dict() == {'Theta_0': {('A', '0', '1'): 1.0}}

        # many rows, with conversions
        marginal_params = pd.DataFrame(
            [
                [np.nan, 1.0, np.nan, np.nan, np.nan, np.nan],
                ['normal', np.nan, 1.0, np.nan, -0.50, 0.50],
                ['lognormal', 1.0, 0.5, np.nan, 0.50, 1.50],
                ['uniform', 0.0, 10.0, np.nan, np.nan, np.nan],
            ],
            columns=[
                'Family',
                'Theta_0',
                'Theta_1',
                'Theta_2',
                'TruncateLower',
                'TruncateUpper',
            ],
            index=pd.MultiIndex.from_tuples(
                (
                    ('A', '0', '1'),
                    ('B', '0', '1'),
                    ('C', '0', '1'),
                    ('D', '0', '1'),
                ),
                names=('cmp', 'loc', 'dir'),
            ),
        )
        units = pd.Series(['ea', 'ft', 'in', 'in2'], index=marginal_params.index)
        arg_units = None
        res = pelicun_model._convert_marginal_params(
            marginal_params, units, arg_units
        )

        expected_df = pd.DataFrame(
            {
                'Family': [np.nan, 'normal', 'lognormal', 'uniform'],
                'Theta_0': [1.0000, np.nan, 0.0254, 0.0000],
                'Theta_1': [np.nan, 1.000000, 0.500000, 0.0064516],
                'Theta_2': [np.nan, np.nan, np.nan, np.nan],
                'TruncateLower': [np.nan, -0.1524, 0.0127, np.nan],
                'TruncateUpper': [np.nan, 0.1524, 0.0381, np.nan],
            },
            index=pd.MultiIndex.from_tuples(
                (
                    ('A', '0', '1'),
                    ('B', '0', '1'),
                    ('C', '0', '1'),
                    ('D', '0', '1'),
                ),
                names=('cmp', 'loc', 'dir'),
            ),
        )

        pd.testing.assert_frame_equal(
            expected_df, res, check_index_type=False, check_column_type=False
        )

        # a case with arg_units
        marginal_params = pd.DataFrame(
            [['500.0,400.00|20,10']],
            columns=['Theta_0'],
            index=pd.MultiIndex.from_tuples(
                (('A', '0', '1'),), names=('cmp', 'loc', 'dir')
            ),
        )
        units = pd.Series(['test_three'], index=marginal_params.index)
        arg_units = pd.Series(['test_two'], index=marginal_params.index)
        res = pelicun_model._convert_marginal_params(
            marginal_params, units, arg_units
        )

        # >>> res
        #                              Theta_0
        # cmp loc dir
        # A   0   1    750,600|40,20

        # note: '40,20' = '20,10' * 2.00 (test_two)
        # note: '750,600' = '500,400' * 3.00 / 2.00 (test_three/test_two)

        expected_df = pd.DataFrame(
            {
                'Theta_0': ['750,600|40,20'],
            },
            index=pd.MultiIndex.from_tuples(
                (('A', '0', '1'),),
                names=('cmp', 'loc', 'dir'),
            ),
        )
        pd.testing.assert_frame_equal(
            expected_df, res, check_index_type=False, check_column_type=False
        )

        # a case with arg_units where we don't divide
        marginal_params = pd.DataFrame(
            [['1.00,2.00|1.00,4.00']],
            columns=['Theta_0'],
            index=[1],
        )
        units = pd.Series(['test_three'], index=marginal_params.index)
        arg_units = pd.Series(['test_two'], index=marginal_params.index)
        res = pelicun_model._convert_marginal_params(
            marginal_params, units, arg_units, divide_units=False
        )

        #    Theta_0
        # 1  3,6|2,8

        # note: '3,6' = '1,2' * 3.00 (test_three)
        # note: '2,8' = '1,4' * 2.00 (test_two)

        expected_df = pd.DataFrame(
            {
                'Theta_0': ['3,6|2,8'],
            },
            index=[1],
        )
        pd.testing.assert_frame_equal(
            expected_df, res, check_index_type=False, check_column_type=False
        )

    def test_query_error_setup(self, pelicun_model: PelicunModel) -> None:
        assert (
            pelicun_model.query_error_setup(
                'Loss/ReplacementThreshold/RaiseOnUnknownKeys'
            )
            is True
        )
        with pytest.raises(KeyError):
            pelicun_model.query_error_setup('some/invalid/path')



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California

# This file is part of pelicun.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

"""Pelicun models."""

from __future__ import annotations

from pelicun.model.asset_model import AssetModel
from pelicun.model.damage_model import DamageModel, DamageModel_DS
from pelicun.model.demand_model import DemandModel
from pelicun.model.loss_model import (
    LossModel,
    RepairModel_DS,
    RepairModel_LF,
)
from pelicun.model.pelicun_model import PelicunModel



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis


"""DemandModel object and associated methods."""

from __future__ import annotations

import re
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, overload

import numexpr as ne
import numpy as np
import pandas as pd

from pelicun import base, file_io, uq
from pelicun.model.pelicun_model import PelicunModel

if TYPE_CHECKING:
    from pelicun.assessment import AssessmentBase

idx = base.idx


class DemandModel(PelicunModel):
    """
    Manages demand information used in assessments.

    Parameters
    ----------
    marginal_params: DataFrame
        Available after the model has been calibrated or calibration data has
        been imported. Defines the marginal distribution of each demand
        variable.
    correlation: DataFrame
        Available after the model has been calibrated or calibration data has
        been imported. Defines the correlation between the demand variables in
        standard normal space. That is, the variables are sampled in standard
        normal space and then transformed into the space of their respective
        distributions and the correlation matrix corresponds to the space where
        they are sampled.
    empirical_data: DataFrame
        Available after the model has been calibrated or calibration data has
        been imported. It provides an empirical dataset for the demand
        variables that are modeled with an empirical distribution.
    sample: DataFrame
        Available after a sample has been generated. Demand variables are
        listed in columns and each row provides an independent realization of
        the joint demand distribution.
    units: Series
        Available after any demand data has been loaded. The index identifies
        the demand variables and the values provide the unit for each variable.
    calibrated: bool
        Signifies whether the DemandModel object has been calibrated.

    """

    __slots__ = [
        '_RVs',
        'calibrated',
        'correlation',
        'empirical_data',
        'marginal_params',
        'sample',
        'user_units',
    ]

    def __init__(self, assessment: AssessmentBase) -> None:
        """
        Instantiate a DemandModel.

        Parameters
        ----------
        assessment: Assessment
            Parent assessment object.

        """
        super().__init__(assessment)

        self.marginal_params: pd.DataFrame | None = None
        self.correlation: pd.DataFrame | None = None
        self.empirical_data: pd.DataFrame | None = None
        self.user_units: pd.Series | None = None
        self.calibrated = False

        self._RVs: uq.RandomVariableRegistry | None = None
        self.sample: pd.DataFrame | None = None

    @overload
    def save_sample(
        self, filepath: None = None, *, save_units: bool = False
    ) -> tuple[pd.DataFrame, pd.Series] | pd.DataFrame: ...

    @overload
    def save_sample(self, filepath: str, *, save_units: bool = False) -> None: ...

    def save_sample(
        self, filepath: str | None = None, *, save_units: bool = False
    ) -> None | tuple[pd.DataFrame, pd.Series] | pd.DataFrame:
        """
        Save demand sample to a csv file or return it in a DataFrame.

        Returns
        -------
        None or tuple
            If `filepath` is specified, the function saves the demand
            sample to a CSV file and returns None.
            If `filepath` is not specified, it returns the DataFrame
            containing the demand sample.
            If `save_units` is True, it returns a tuple of the
            DataFrame and a Series containing the units.

        """
        self.log.div()
        if filepath is not None:
            self.log.msg('Saving demand sample...')

        assert self.sample is not None
        res = file_io.save_to_csv(
            self.sample,
            Path(filepath) if filepath is not None else None,
            units=self.user_units,
            unit_conversion_factors=self._asmnt.unit_conversion_factors,
            use_simpleindex=(filepath is not None),
            log=self._asmnt.log,
        )
        if filepath is not None:
            self.log.msg(
                'Demand sample successfully saved.', prepend_timestamp=False
            )
            return None

        # else:
        assert isinstance(res, pd.DataFrame)

        units = res.loc['Units']
        res = res.drop('Units')
        assert isinstance(units, pd.Series)

        if save_units:
            return res.astype(float), units

        # else:
        return res.astype(float)

    def load_sample(self, filepath: str | pd.DataFrame) -> None:
        """
        Load demand sample data and parse it.

        Besides parsing the sample, the method also reads and saves
        the units specified for each demand variable. If no units are
        specified, base units are assumed.

        Parameters
        ----------
        filepath: string or DataFrame
            Location of the file with the demand sample.

        """

        def parse_header(raw_header: pd.Index[str]) -> pd.Index[str]:
            """
            Parse and clean header.

            Parses and cleans the header of a demand DataFrame from
            raw multi-level index to a standardized format.

            This function adjusts the raw header of a DataFrame,
            removing optional event IDs and whitespace, and
            standardizing the format to facilitate further
            processing. It is designed to handle headers with either
            three or four levels, where the first level (event_ID) is
            optional and not used in further analysis.  Note: This
            will soon change, and that first level will be enforced
            instead of removed.

            Parameters
            ----------
            raw_header: pd.MultiIndex
                The original multi-level index (header) of the
                DataFrame, which may contain an optional event_ID and
                might have excess whitespace in the labels.

            Returns
            -------
            pd.MultiIndex
                A new MultiIndex for the DataFrame's columns that is
                cleaned of any unwanted characters or levels. This
                index has three levels: 'type', 'loc', and 'dir',
                representing the type of demand, location, and
                direction, respectively.

            """
            old_mi = raw_header

            # The first number (event_ID) in the demand labels is optional and
            # currently not used. We remove it if it was in the raw data.
            num_levels_with_event_id = 4
            if old_mi.nlevels == num_levels_with_event_id:
                if self._asmnt.log.verbose:
                    self.log.msg(
                        'Removing event_ID from header...', prepend_timestamp=False
                    )

                new_column_index_array = np.array(
                    [old_mi.get_level_values(i) for i in range(1, 4)]
                )

            else:
                new_column_index_array = np.array(
                    [old_mi.get_level_values(i) for i in range(3)]
                )

            # Remove whitespace to avoid ambiguity

            if self._asmnt.log.verbose:
                self.log.msg(
                    'Removing whitespace from header...', prepend_timestamp=False
                )

            wspace_remove = np.vectorize(lambda name: str(name).replace(' ', ''))

            new_column_index = wspace_remove(new_column_index_array)

            # Creating new, cleaned-up header

            return pd.MultiIndex.from_arrays(
                new_column_index, names=['type', 'loc', 'dir']
            )

        self.log.div()
        self.log.msg('Loading demand data...')

        demand_data, units = file_io.load_data(
            filepath,
            self._asmnt.unit_conversion_factors,
            return_units=True,
            log=self._asmnt.log,
        )
        assert isinstance(demand_data, pd.DataFrame)
        assert isinstance(units, pd.Series)

        parsed_data = demand_data.copy()

        # start with cleaning up the header

        parsed_data.columns = parse_header(parsed_data.columns)

        # Remove errors, if needed
        if 'ERROR' in parsed_data.columns.get_level_values(0):
            self.log.msg(
                'Removing errors from the raw data...', prepend_timestamp=False
            )

            error_list = (
                parsed_data.loc[  # type: ignore
                    :,  # type: ignore
                    idx['ERROR', :, :],  # type: ignore
                ]
                .to_numpy()
                .astype(  # type: ignore
                    bool  # type: ignore
                )
            )  # type: ignore

            parsed_data = parsed_data.loc[~error_list, :].copy()
            parsed_data = parsed_data.drop('ERROR', level=0, axis=1)

            self.log.msg(
                '\nBased on the values in the ERROR column, '
                f'{np.sum(error_list)} demand samples were removed.\n',
                prepend_timestamp=False,
            )

        self.sample = parsed_data

        self.log.msg('Demand data successfully parsed.', prepend_timestamp=False)

        # parse the index for the units
        units.index = parse_header(units.index)

        self.user_units = units

        self.log.msg('Demand units successfully parsed.', prepend_timestamp=False)

    def estimate_RID(  # noqa: N802
        self,
        demands: pd.DataFrame | pd.Series,
        params: dict,
        method: str = 'FEMA P58',
    ) -> pd.DataFrame:
        """
        Estimate residual inter-story drift (RID).

        Estimates residual inter-story drift (RID) realizations based
        on peak inter-story drift (PID) and other demand parameters
        using specified methods.

        This method calculates RID based on the peak inter-story drift
        provided in the demands DataFrame and parameters such as yield
        drift specified in the params dictionary. The calculation
        adheres to the FEMA P-58 methodology, which includes
        conditions for different ranges of drift.

        Parameters
        ----------
        demands: DataFrame
            A DataFrame containing samples of demands, specifically
            peak inter-story drift (PID) values for various
            location-direction pairs required for the estimation
            method.
        params: dict
            A dictionary containing parameters required for the
            estimation method, such as 'yield_drift', which is the
            drift at which yielding is expected to occur.
        method: str, optional
            The method used to estimate the RID values. Currently,
            only 'FEMA P58' is implemented. Defaults to 'FEMA P58'.

        Returns
        -------
        DataFrame
            A DataFrame containing the estimated residual inter-story
            drift (RID) realizations, indexed and structured similarly
            to the input demands DataFrame.

        Raises
        ------
        ValueError
            Raises a ValueError if an unrecognized method is provided
            or required parameters are missing in the `params`
            dictionary.

        Notes
        -----
        The FEMA P-58 estimation approach divides the drift into three
        domains, with different transformation rules for
        each. Additional stochastic variation is introduced to nonzero
        RID values to model the inherent uncertainty. The method
        ensures that the RID values do not exceed the corresponding
        PID values.

        """
        if method in {'FEMA P58', 'FEMA P-58'}:
            # method is described in FEMA P-58 Volume 1 Section 5.4 &
            # Appendix C

            # the provided demands shall be PID values at various
            # loc-dir pairs
            pid = demands

            # there's only one parameter needed: the yield drift
            yield_drift = params['yield_drift']

            # three subdomains of demands are identified
            small = yield_drift > pid
            medium = 4 * yield_drift > pid
            large = 4 * yield_drift <= pid

            # convert PID to RID in each subdomain
            rid = pid.copy()
            rid[large] = pid[large] - 3 * yield_drift
            rid[medium] = 0.3 * (pid[medium] - yield_drift)
            rid[small] = 0.0

            # add extra uncertainty to nonzero values
            rng = self._asmnt.options.rng
            eps = rng.normal(scale=0.2, size=rid.shape)
            rid[rid > 0] = np.exp(np.log(rid[rid > 0]) + eps)  # type: ignore

            # finally, make sure the RID values are never larger than
            # the PIDs
            rid = pd.DataFrame(
                np.minimum(pid.values, rid.values),  # type: ignore
                columns=pd.DataFrame(  # noqa: PD013
                    1,
                    index=['RID'],
                    columns=pid.columns,
                )
                .stack(level=[0, 1])
                .index,
                index=pid.index,
            )

        else:
            msg = f'Invalid method: `{method}`.'
            raise ValueError(msg)

        return rid

    def estimate_RID_and_adjust_sample(  # noqa: N802
        self, params: dict, method: str = 'FEMA P58'
    ) -> None:
        """
        Estimate residual inter-story drift (RID) and modifies sample.

        Uses `self.estimate_RID` and adjusts the demand sample.
        See the docstring of the `estimate_RID` method for details.

        Parameters
        ----------
        params: dict
            A dictionary containing parameters required for the
            estimation method, such as 'yield_drift', which is the
            drift at which yielding is expected to occur.
        method: str, optional
            The method used to estimate the RID values. Currently,
            only 'FEMA P58' is implemented. Defaults to 'FEMA P58'.

        Raises
        ------
        ValueError
            If the method is called before a sample is generated.

        """
        if self.sample is None:
            msg = 'Demand model does not have a sample yet.'
            raise ValueError(msg)

        sample_tuple = self.save_sample(save_units=True)
        assert isinstance(sample_tuple, tuple)
        demand_sample, demand_units = sample_tuple
        assert isinstance(demand_sample, pd.DataFrame)
        assert isinstance(demand_units, pd.Series)
        pid = demand_sample['PID']
        rid = self.estimate_RID(pid, params, method)
        rid_units = pd.Series('unitless', index=rid.columns)
        demand_sample_ext = pd.concat([demand_sample, rid], axis=1)
        units_ext = pd.concat([demand_units, rid_units])
        demand_sample_ext.loc['Units', :] = units_ext
        self.load_sample(demand_sample_ext)

    def expand_sample(
        self,
        label: str,
        value: float | np.ndarray,
        unit: str,
        location: str = '0',
        direction: str = '1',
    ) -> None:
        """
        Add an extra column to the demand sample.

        The column contains repeated instances of `value`, is accessed
        via the multi-index (`label`-`location`-`direction`), and has
        units of `unit`.

        Parameters
        ----------
        label: str
            Label to use to extend the MultiIndex of the demand sample.
        value: float | np.ndarray
            Values to add to the rows of the additional column.
        unit: str
            Unit that corresponds to the additional column.
        location: str, optional
            Optional location, defaults to `0`.
        direction: str, optional
            Optional direction, defaults to `1`.

        Raises
        ------
        ValueError
            If the method is called before a sample is generated.
        ValueError
            If `value` is a numpy array of incorrect shape.

        """
        if self.sample is None:
            msg = 'Demand model does not have a sample yet.'
            raise ValueError(msg)
        sample_tuple = self.save_sample(save_units=True)
        assert isinstance(sample_tuple, tuple)
        demand_sample, demand_units = sample_tuple
        assert isinstance(demand_sample, pd.DataFrame)
        assert isinstance(demand_units, pd.Series)
        if isinstance(value, np.ndarray) and len(value) != len(demand_sample):
            msg = 'Incompatible array length.'
            raise ValueError(msg)
        demand_sample[label, location, direction] = value
        demand_units[label, location, direction] = unit
        demand_sample.loc['Units', :] = demand_units
        self.load_sample(demand_sample)

    def calibrate_model(self, config: dict) -> None:  # noqa: C901
        """
        Calibrate a demand model to describe the raw demand data.

        The raw data shall be parsed first to ensure that it follows the
        schema expected by this method. The calibration settings define the
        characteristics of the multivariate distribution that is fit to the
        raw data.

        Parameters
        ----------
        config: dict
            A dictionary, typically read from a JSON file, that specifies the
            distribution family, truncation and censoring limits, and other
            settings for the calibration.

        """
        if self.calibrated:
            self.log.warning('DemandModel has been previously calibrated.')

        def parse_settings(  # noqa: C901
            cal_df: pd.DataFrame, settings: dict, demand_type: str
        ) -> None:
            def parse_str_to_float(in_str: str, context_string: str) -> float:
                try:
                    out_float = (
                        np.nan if base.check_if_str_is_na(in_str) else float(in_str)
                    )

                except ValueError:
                    self.log.warning(
                        f'Could not parse {in_str} provided as '
                        f'{context_string}. Using NaN instead.'
                    )

                    out_float = np.nan

                return out_float

            demand_sample = self.save_sample()
            assert isinstance(demand_sample, pd.DataFrame)
            active_d_types = demand_sample.columns.get_level_values('type').unique()

            if demand_type == 'ALL':
                cols = tuple(active_d_types)

            else:
                cols_lst = []

                for d_type in active_d_types:
                    if d_type.split('_')[0] == demand_type:
                        cols_lst.append(d_type)  # noqa: PERF401

                cols = tuple(cols_lst)

            # load the distribution family
            cal_df.loc[list(cols), 'Family'] = settings['DistributionFamily']

            # load limits
            for lim in (
                'CensorLower',
                'CensorUpper',
                'TruncateLower',
                'TruncateUpper',
            ):
                if lim in settings:
                    val = parse_str_to_float(settings[lim], lim)
                    if not pd.isna(val):
                        cal_df.loc[list(cols), lim] = val

            # scale the censor and truncation limits, if needed
            scale_factor = self._asmnt.scale_factor(settings.get('Unit'))

            rows_to_scale = [
                'CensorLower',
                'CensorUpper',
                'TruncateLower',
                'TruncateUpper',
            ]
            cal_df.loc[idx[cols, :, :], rows_to_scale] *= scale_factor  # type: ignore

            # load the prescribed additional uncertainty
            if 'AddUncertainty' in settings:
                sig_increase = parse_str_to_float(
                    settings['AddUncertainty'], 'AddUncertainty'
                )

                # scale the sig value if the target distribution family is normal
                if settings['DistributionFamily'] == 'normal':
                    sig_increase *= scale_factor

                cal_df.loc[list(cols), 'SigIncrease'] = sig_increase

        def get_filter_mask(
            demand_sample: pd.DataFrame,
            lower_lims: np.ndarray,
            upper_lims: np.ndarray,
        ) -> bool:
            demands_of_interest = demand_sample.iloc[:, pd.notna(upper_lims)]
            limits_of_interest = upper_lims[pd.notna(upper_lims)]
            upper_mask = np.all(demands_of_interest < limits_of_interest, axis=1)

            demands_of_interest = demand_sample.iloc[:, pd.notna(lower_lims)]
            limits_of_interest = lower_lims[pd.notna(lower_lims)]
            lower_mask = np.all(demands_of_interest > limits_of_interest, axis=1)

            return np.all([lower_mask, upper_mask], axis=0)

        self.log.div()
        self.log.msg('Calibrating demand model...')

        demand_sample = self.sample
        assert isinstance(demand_sample, pd.DataFrame)

        # initialize a DataFrame that contains calibration information
        cal_df = pd.DataFrame(
            columns=[
                'Family',
                'CensorLower',
                'CensorUpper',
                'TruncateLower',
                'TruncateUpper',
                'SigIncrease',
                'Theta_0',
                'Theta_1',
            ],
            index=demand_sample.columns,
            dtype=float,
        )

        cal_df['Family'] = cal_df['Family'].astype(str)

        # start by assigning the default option ('ALL') to every demand column
        parse_settings(cal_df, config['ALL'], demand_type='ALL')

        # then parse the additional settings and make the necessary adjustments
        for demand_type in config:  # noqa: PLC0206
            if demand_type != 'ALL':
                parse_settings(cal_df, config[demand_type], demand_type=demand_type)

        if self._asmnt.log.verbose:
            self.log.msg(
                '\nCalibration settings successfully parsed:\n' + str(cal_df),
                prepend_timestamp=False,
            )
        else:
            self.log.msg(
                '\nCalibration settings successfully parsed:\n',
                prepend_timestamp=False,
            )

        # save the settings
        model_params = cal_df.copy()

        # Remove the samples outside of censoring limits
        # Currently, non-empirical demands are assumed to have some level of
        # correlation, hence, a censored value in any demand triggers the
        # removal of the entire sample from the population.
        upper_lims = cal_df.loc[:, 'CensorUpper'].to_numpy()
        lower_lims = cal_df.loc[:, 'CensorLower'].to_numpy()

        assert isinstance(demand_sample, pd.DataFrame)
        if ~np.all(pd.isna(np.array([upper_lims, lower_lims]))):
            censor_mask = get_filter_mask(demand_sample, lower_lims, upper_lims)
            censored_count = np.sum(~censor_mask)

            demand_sample = pd.DataFrame(demand_sample.loc[censor_mask, :])

            self.log.msg(
                '\nBased on the provided censoring limits, '
                f'{censored_count} samples were censored.',
                prepend_timestamp=False,
            )
        else:
            censored_count = 0

        # Check if there is any sample outside of truncation limits
        # If yes, that suggests an error either in the samples or the
        # configuration. We handle such errors gracefully: the analysis is not
        # terminated, but we show an error in the log file.
        upper_lims = cal_df.loc[:, 'TruncateUpper'].to_numpy()
        lower_lims = cal_df.loc[:, 'TruncateLower'].to_numpy()

        assert isinstance(demand_sample, pd.DataFrame)
        if ~np.all(pd.isna(np.array([upper_lims, lower_lims]))):
            truncate_mask = get_filter_mask(demand_sample, lower_lims, upper_lims)
            truncated_count = np.sum(~truncate_mask)

            if truncated_count > 0:
                demand_sample = pd.DataFrame(demand_sample.loc[truncate_mask, :])

                self.log.msg(
                    '\nBased on the provided truncation limits, '
                    f'{truncated_count} samples were removed before demand '
                    'calibration.',
                    prepend_timestamp=False,
                )

        # Separate and save the demands that are kept empirical -> i.e., no
        # fitting. Currently, empirical demands are decoupled from those that
        # have a distribution fit to their samples. The correlation between
        # empirical and other demands is not preserved in the demand model.
        empirical_edps = []
        for edp in cal_df.index:
            if cal_df.loc[edp, 'Family'] == 'empirical':
                empirical_edps.append(edp)  # noqa: PERF401

        assert isinstance(demand_sample, pd.DataFrame)
        if empirical_edps:
            self.empirical_data = demand_sample.loc[:, empirical_edps].copy()

        # remove the empirical demands from the samples used for calibration
        demand_sample = demand_sample.drop(empirical_edps, axis=1)

        # and the calibration settings
        cal_df = cal_df.drop(empirical_edps, axis=0)

        if self._asmnt.log.verbose:
            self.log.msg(
                f'\nDemand data used for calibration:\n{demand_sample}',
                prepend_timestamp=False,
            )

        # fit the joint distribution
        self.log.msg(
            '\nFitting the prescribed joint demand distribution...',
            prepend_timestamp=False,
        )

        demand_theta, demand_rho = uq.fit_distribution_to_sample(
            raw_sample=demand_sample.to_numpy().T,
            distribution=cal_df.loc[:, 'Family'].values,  # type: ignore
            censored_count=censored_count,
            detection_limits=cal_df.loc[  # type: ignore
                :,
                ['CensorLower', 'CensorUpper'],
            ].values,
            truncation_limits=cal_df.loc[  # type: ignore
                :, ['TruncateLower', 'TruncateUpper']
            ].values,
            multi_fit=False,
            logger_object=self._asmnt.log,
        )
        # fit the joint distribution
        self.log.msg(
            '\nCalibration successful, processing results...',
            prepend_timestamp=False,
        )

        # save the calibration results
        model_params.loc[cal_df.index, ['Theta_0', 'Theta_1']] = demand_theta

        # increase the variance of the marginal distributions, if needed
        if ~np.all(pd.isna(model_params.loc[:, 'SigIncrease'].values)):
            self.log.msg('\nIncreasing demand variance...', prepend_timestamp=False)

            sig_inc = np.nan_to_num(
                model_params.loc[:, 'SigIncrease'].values,  # type: ignore
            )
            sig_0 = model_params.loc[:, 'Theta_1'].to_numpy()

            model_params.loc[:, 'Theta_1'] = np.sqrt(
                sig_0**2.0 + sig_inc**2.0,  # type: ignore
            )

        # remove unneeded fields from model_params
        for col in ('SigIncrease', 'CensorLower', 'CensorUpper'):
            model_params = model_params.drop(col, axis=1)

        # reorder the remaining fields for clarity
        model_params = model_params[
            ['Family', 'Theta_0', 'Theta_1', 'TruncateLower', 'TruncateUpper']
        ]

        self.marginal_params = model_params

        self.log.msg(
            '\nCalibrated demand model marginal distributions:\n'
            + str(model_params),
            prepend_timestamp=False,
        )

        # save the correlation matrix
        self.correlation = pd.DataFrame(
            demand_rho, columns=cal_df.index, index=cal_df.index
        )

        self.log.msg(
            '\nCalibrated demand model correlation matrix:\n'
            + str(self.correlation),
            prepend_timestamp=False,
        )

        self.calibrated = True

    def save_model(self, file_prefix: str) -> None:
        """Save parameters of the demand model to a set of csv files."""
        self.log.div()
        self.log.msg('Saving demand model...')

        # save the correlation and empirical data
        file_io.save_to_csv(self.correlation, Path(file_prefix + '_correlation.csv'))
        if self.empirical_data is not None:
            file_io.save_to_csv(
                self.empirical_data,
                Path(file_prefix + '_empirical.csv'),
                units=self.user_units,
                unit_conversion_factors=self._asmnt.unit_conversion_factors,
                log=self._asmnt.log,
            )

        # Converting the marginal parameters requires special
        # treatment, so we can't rely on file_io's universal unit
        # conversion functionality. We do it manually here instead.
        assert isinstance(self.marginal_params, pd.DataFrame)
        assert isinstance(self.user_units, pd.Series)
        marginal_params_user_units = self._convert_marginal_params(
            self.marginal_params.copy(), self.user_units, inverse_conversion=True
        )
        marginal_params_user_units['Units'] = self.user_units

        file_io.save_to_csv(
            marginal_params_user_units,
            Path(file_prefix + '_marginals.csv'),
            orientation=1,
            log=self._asmnt.log,
        )

        self.log.msg('Demand model successfully saved.', prepend_timestamp=False)

    def load_model(self, data_source: str | dict) -> None:
        """
        Load the model that describes demands on the asset.

        Parameters
        ----------
        data_source: string or dict
            If string, the data_source is a file prefix (<prefix> in the
            following description) that identifies the following files:
            <prefix>_marginals.csv,  <prefix>_empirical.csv,
            <prefix>_correlation.csv. If dict, the data source is a dictionary
            with the following optional keys: 'marginals', 'empirical', and
            'correlation'. The value under each key shall be a DataFrame.

        Raises
        ------
        TypeError
            If the data source type is invalid.

        """
        self.log.div()
        self.log.msg('Loading demand model...')

        # prepare the marginal data source variable to load the data
        if isinstance(data_source, dict):
            marginal_data_source = data_source.get('marginals')
            assert isinstance(marginal_data_source, pd.DataFrame)
            empirical_data_source = data_source.get('empirical', None)
            correlation_data_source = data_source.get('correlation', None)
        elif isinstance(data_source, str):
            marginal_data_source = data_source + '_marginals.csv'
            empirical_data_source = data_source + '_empirical.csv'
            correlation_data_source = data_source + '_correlation.csv'
        else:
            msg = f'Invalid data_source type: {type(data_source)}.'
            raise TypeError(msg)

        if empirical_data_source is not None:
            if (
                isinstance(empirical_data_source, str)
                and Path(empirical_data_source).exists()
            ):
                empirical_data = file_io.load_data(
                    empirical_data_source,
                    self._asmnt.unit_conversion_factors,
                    log=self._asmnt.log,
                )
                assert isinstance(empirical_data, pd.DataFrame)
                self.empirical_data = empirical_data
                self.empirical_data.columns.names = ['type', 'loc', 'dir']
        else:
            self.empirical_data = None

        if correlation_data_source is not None:
            correlation = file_io.load_data(
                correlation_data_source,
                self._asmnt.unit_conversion_factors,
                reindex=False,
                log=self._asmnt.log,
            )
            assert isinstance(correlation, pd.DataFrame)
            self.correlation = correlation
            self.correlation.index = self.correlation.index.set_names(
                ['type', 'loc', 'dir']
            )
            self.correlation.columns = self.correlation.columns.set_names(
                ['type', 'loc', 'dir']
            )
        else:
            self.correlation = None

        # Note that a data source without marginal information is not valid
        marginal_params, units = file_io.load_data(
            marginal_data_source,
            None,
            orientation=1,
            reindex=False,
            return_units=True,
            log=self._asmnt.log,
        )
        assert isinstance(marginal_params, pd.DataFrame)
        assert isinstance(units, pd.Series)

        marginal_params.index = marginal_params.index.set_names(
            ['type', 'loc', 'dir']
        )

        marginal_params = self._convert_marginal_params(
            marginal_params.copy(), units
        )

        self.marginal_params = marginal_params
        self.user_units = units

        self.log.msg('Demand model successfully loaded.', prepend_timestamp=False)

    def _create_RVs(self, *, preserve_order: bool = False) -> None:  # noqa: N802
        """Create a random variable registry for the joint distribution of demands."""
        assert self.marginal_params is not None

        # initialize the registry
        rv_reg = uq.RandomVariableRegistry(self._asmnt.options.rng)

        # add a random variable for each demand variable
        for rv_params in self.marginal_params.itertuples():
            edp = rv_params.Index
            rv_tag = f'EDP-{edp[0]}-{edp[1]}-{edp[2]}'  # type: ignore
            family = getattr(rv_params, 'Family', 'deterministic')

            if family == 'empirical':
                dist_family = 'coupled_empirical' if preserve_order else 'empirical'

                # empirical RVs need the data points
                rv_reg.add_RV(
                    uq.rv_class_map(dist_family)(
                        name=rv_tag,
                        theta=self.empirical_data.loc[  # type: ignore
                            :,  # type: ignore
                            edp,
                        ].values,
                    )
                )

            else:
                # all other RVs need parameters of their distributions
                rv_reg.add_RV(
                    uq.rv_class_map(family)(  # type: ignore
                        name=rv_tag,
                        theta=np.array(
                            [
                                value
                                for t_i in range(3)
                                if (
                                    value := getattr(rv_params, f'Theta_{t_i}', None)
                                )
                                is not None
                            ]
                        ),
                        truncation_limits=np.array(
                            [
                                getattr(rv_params, f'Truncate{side}', np.nan)
                                for side in ('Lower', 'Upper')
                            ]
                        ),
                    )
                )

        self.log.msg(
            f'\n{self.marginal_params.shape[0]} random variables created.',
            prepend_timestamp=False,
        )

        # add an RV set to consider the correlation between demands, if needed
        if self.correlation is not None:
            rv_set_tags = [
                f'EDP-{edp[0]}-{edp[1]}-{edp[2]}'
                for edp in self.correlation.index.to_numpy()
            ]

            rv_reg.add_RV_set(
                uq.RandomVariableSet(
                    'EDP_set',
                    list(rv_reg.RVs(rv_set_tags).values()),
                    self.correlation.values,
                )
            )

            self.log.msg(
                f'\nCorrelations between {len(rv_set_tags)} random variables '
                'successfully defined.',
                prepend_timestamp=False,
            )

        self._RVs = rv_reg

    def clone_demands(self, demand_cloning: dict) -> None:
        """
        Clone demands.

        Copies over columns of the original demand sample and
        assigns given names to them. The columns to be copied over
        and the names to be assigned to the copies are defined as the keys
        and values of the `demand_cloning` dictionary.
        The method modifies `sample` inplace.

        Parameters
        ----------
        demand_cloning: dict
            Keys correspond to the columns of the original sample to
            be copied over and the values correspond to the intended
            names for the copies. Caution: It's possible to define a
            dictionary with duplicate keys, and Python will just keep
            the last entry without warning. Users need to be careful
            enough to avoid duplicate keys, because we can't validate
            them.
            E.g.: x = {'1': 1.00, '1': 2.00} results in x={'1': 2.00}.

        Raises
        ------
        ValueError
            In multiple instances of invalid demand_cloning entries.

        """
        # it's impossible to have duplicate keys, because
        # demand_cloning is a dictionary.
        new_columns_list = demand_cloning.values()
        # The following prevents duplicate entries in the values
        # corresponding to a single cloned demand (1), but
        # also the same column being specified as the cloned
        # entry of multiple demands (2).
        # e.g.
        # (1): {'PGV-0-1': ['PGV-1-1', 'PGV-1-1', ...]}
        # (2): {'PGV-0-1': ['PGV-1-1', ...], 'PGV-0-2': ['PGV-1-1', ...]}
        flat_list = []
        for new_columns in new_columns_list:
            flat_list.extend(new_columns)
        if len(set(flat_list)) != len(flat_list):
            msg = 'Duplicate entries in demand cloning configuration.'
            raise ValueError(msg)

        # turn the config entries to tuples
        def turn_to_tuples(demand_cloning: dict) -> dict:
            demand_cloning_tuples = {}
            for key, values in demand_cloning.items():
                demand_cloning_tuples[tuple(key.split('-'))] = [
                    tuple(x.split('-')) for x in values
                ]
            return demand_cloning_tuples

        demand_cloning = turn_to_tuples(demand_cloning)

        # The demand cloning configuration should not include
        # columns that are not present in the original sample.
        warn_columns = []
        assert self.sample is not None
        for column in demand_cloning:
            if column not in self.sample.columns:
                warn_columns.append(column)  # noqa: PERF401
        if warn_columns:
            warn_columns = ['-'.join(x) for x in warn_columns]
            self.log.warning(
                'The demand cloning configuration lists '
                "columns that are not present in the original demand sample's "
                f'columns: {warn_columns}.'
            )

        # we iterate over the existing columns of the sample and try
        # to locate columns that need to be copied as required by the
        # demand cloning configuration.  If a column does not need
        # to be cloned it is left as is.  Otherwise, we keep track
        # of its initial index location (in `column_index`) and the
        # number of times it needs to be replicated, along with the
        # new names of its copies (in `column_values`).
        column_index = []
        column_values = []
        for i, column in enumerate(self.sample.columns):
            if column not in demand_cloning:
                column_index.append(i)
                column_values.append(column)
            else:
                new_column_values = demand_cloning[column]
                column_index.extend([i] * len(new_column_values))
                column_values.extend(new_column_values)
        # copy the columns
        self.sample = self.sample.iloc[:, column_index]
        # update the column index
        self.sample.columns = pd.MultiIndex.from_tuples(column_values)
        # update units
        self.user_units = self.user_units.iloc[column_index]  # type: ignore
        assert self.user_units is not None
        self.user_units.index = self.sample.columns

    def generate_sample(self, config: dict) -> None:
        """
        Generate the demand sample.

        Generates a sample of random variables (RVs) based on the
        specified configuration for demand modeling.

        This method utilizes the current settings for marginal
        distribution parameters to generate a sample of demand
        variables. The configuration can specify details such as the
        sample size, whether to preserve the order of raw data, and
        whether to apply demand cloning. The generated sample is
        stored internally and can be used for subsequent analysis.

        Parameters
        ----------
        config: dict
            A dictionary containing configuration options for the
            sample generation. Key options include:
            * 'SampleSize': The number of samples to generate.
            * 'PreserveRawOrder': Boolean indicating whether to
            preserve the order of the raw data. Defaults to False.
            * 'DemandCloning': Specifies if and how demand cloning
            should be applied. Can be a boolean or a detailed
            configuration.

        Raises
        ------
        ValueError
            If model parameters are not loaded or specified before
            attempting to generate a sample.

        Notes
        -----
        The function is responsible for the creation of random
        variables based on the distribution parameters specified in
        `marginal_params`.  It ensures that the sample is properly
        indexed and sorted according to type, location, and
        direction. It also handles the configuration of demand cloning
        if specified, which is a method to artificially augment the
        variability in the sample based on existing realizations.

        Examples
        --------
        >>> config = {
                'SampleSize': 1000,
                'PreserveRawOrder': True,
                'DemandCloning': False
            }
        >>> model.generate_sample(config)
        # This will generate 1000 realizations of demand variables
        # with the specified configuration.

        """
        if self.marginal_params is None:
            msg = (
                'Model parameters have not been specified. Either '
                'load parameters from a file or calibrate the '
                'model using raw demand data.'
            )
            raise ValueError(msg)

        self.log.div()
        self.log.msg('Generating sample from demand variables...')

        self._create_RVs(preserve_order=config.get('PreserveRawOrder', False))

        assert self._RVs is not None
        assert self._asmnt.options.sampling_method is not None
        sample_size = config['SampleSize']
        self._RVs.generate_sample(
            sample_size=sample_size, method=self._asmnt.options.sampling_method
        )

        # replace the potentially existing raw sample with the generated one
        assert self._RVs is not None
        assert self._RVs.RV_sample is not None
        sample = pd.DataFrame(self._RVs.RV_sample)
        sample = sample.sort_index(axis=0)
        sample = sample.sort_index(axis=1)

        sample_mi = base.convert_to_MultiIndex(sample, axis=1)['EDP']
        assert isinstance(sample_mi, pd.DataFrame)
        sample = sample_mi

        sample.columns.names = ['type', 'loc', 'dir']
        self.sample = sample

        if config.get('DemandCloning', False):
            self.clone_demands(config['DemandCloning'])

        self.log.msg(
            f'\nSuccessfully generated {sample_size} realizations.',
            prepend_timestamp=False,
        )


def _get_required_demand_type(
    model_parameters: pd.DataFrame,
    pgb: pd.DataFrame,
    demand_offset: dict | None = None,
) -> dict:
    """
    Get the required demand type for the components.

    Returns the demand type and its properties required to calculate
    the the damage or loss of a component. The properties include
    whether the demand is directional, the offset, and the type of the
    demand. The method takes as input a dataframe `PGB` that contains
    information about the component groups in the asset. For each
    performance group PG in the PGB dataframe, the method retrieves
    the relevant parameters from the model_params dataframe and parses
    the demand type into its properties. If the demand type has a
    subtype, the method splits it and adds the subtype to the demand
    type to form the EDP type. The method also considers the default
    offset for the demand type, if it is specified in the options
    attribute of the assessment, and adds the offset to the EDP. If
    the demand is directional, the direction is added to the EDP. The
    method collects all the unique EDPs for each component group and
    returns them as a dictionary where each key is an EDP and its
    value is a list of component groups that require that EDP.

    Parameters
    ----------
    model_parameters: pd.DataFrame
        Model parameters. Damage model parameters, or
        loss-function loss model parameters.
    pgb: pd.DataFrame
        A pandas DataFrame with the block information for
        each component
    demand_offset: dict, optional
        Specifies an additional location offset for specific
        demand types. Example:
        {'PFA': -1, 'PFV': +2}.

    Returns
    -------
    dict
        A dictionary of EDP requirements, where each key is the EDP
        string (e.g., "PGA-0-1"), and the
        corresponding value is a list of tuples (component_id,
        location, direction)

    Raises
    ------
    ValueError
        When a negative value is used for `loc`. Currently not
        supported.

    """
    model_parameters = model_parameters.sort_index(axis=1)

    # Assign default demand_offset to empty dict.
    if not demand_offset:
        demand_offset = {}

    required_edps = defaultdict(list)

    for pg in pgb.index:
        cmp = pg[0]

        # Utility Demand: if there is an `Expression`, then load the
        # rest of the demand types.
        expression = model_parameters.loc[cmp, :].get(('Demand', 'Expression'))
        if expression is not None:
            # get the number of variables in the expression using
            # the numexpr library
            parsed_expr = ne.NumExpr(_clean_up_expression(expression))
            num_terms = len(parsed_expr.input_names)
            demand_parameters_list = []
            for i in range(num_terms):
                if i == 0:
                    index_lvl0 = 'Demand'
                else:
                    index_lvl0 = f'Demand{i + 1}'
                demand_parameters_list.append(
                    (
                        model_parameters.loc[cmp, (index_lvl0, 'Type')],
                        model_parameters.loc[cmp, (index_lvl0, 'Offset')],
                        model_parameters.loc[cmp, (index_lvl0, 'Directional')],
                    )
                )
        else:
            demand_parameters_list = [
                (
                    model_parameters.loc[cmp, ('Demand', 'Type')],
                    model_parameters.loc[cmp, ('Demand', 'Offset')],
                    model_parameters.loc[cmp, ('Demand', 'Directional')],
                )
            ]

        # Parse the demand type

        edps = []
        for demand_parameters in demand_parameters_list:
            demand_type = demand_parameters[0]
            offset = demand_parameters[1]
            directional = demand_parameters[2]

            assert isinstance(demand_type, str)

            # Check if there is a subtype included in the demand_type
            # string
            if '|' in demand_type:
                # If there is a subtype, split the demand_type string
                # on the '|' character
                demand_type, subtype = demand_type.split('|')
                # Convert the demand type to the corresponding EDP
                # type using `base.EDP_to_demand_type`
                demand_type = base.EDP_to_demand_type[demand_type]
                # Concatenate the demand type and subtype to form the
                # EDP type
                edp_type = f'{demand_type}_{subtype}'
            else:
                # If there is no subtype, convert the demand type to
                # the corresponding EDP type using
                # `base.EDP_to_demand_type`
                demand_type = base.EDP_to_demand_type[demand_type]
                # Assign the EDP type to be equal to the demand type
                edp_type = demand_type

            # Consider the default offset, if needed
            if demand_type in demand_offset:
                # If the demand type has a default offset in
                # `demand_offset`, add the offset
                # to the default offset
                offset = int(offset + demand_offset[demand_type])  # type: ignore
            else:
                # If the demand type does not have a default offset in
                # `demand_offset`, convert the
                # offset to an integer
                offset = int(offset)  # type: ignore

            # Determine the direction
            direction = pg[2] if directional else '0'

            # Concatenate the EDP type, offset, and direction to form
            # the EDP key
            edp = f'{edp_type}-{int(pg[1]) + offset!s}-{direction}'

            if int(pg[1]) + offset < 0:
                msg = (
                    f'Negative location encountered for component '
                    f'(cmp, loc, dir, uid)=`{pg}`. Would require `{edp}`. '
                    f'Please update the location of the component.'
                )
                raise ValueError(msg)

            edps.append(edp)

        edps_t = tuple(edps)  # makes it hashable

        # Add the current PG (performance group) to the list of
        # PGs associated with the current EDP key
        required_edps[edps_t, expression].append(pg)

    # Return the required EDPs
    return required_edps


def _assemble_required_demand_data(
    required_edps: set, nondirectional_multipliers: dict, demand_sample: pd.DataFrame
) -> dict:
    """
    Assembles demand data for damage state determination.

    The method takes the maximum of all available directions for
    non-directional demand, scaling it using the non-directional
    multiplier specified in self._asmnt.options, and returning the
    result as a dictionary with keys in the format of
    '<demand_type>-<location>-<direction>' and values as arrays of
    demand values.

    Parameters
    ----------
    required_edps: set
        Set of required EDPs. The elements in the set should be
        tuples. For each, the first element should be a tuple
        containing EDPs in the `type`-`loc`-`dir` format. The second
        should be an expression defining how the EDPs in the tuple
        should be combined when it contains more than a single EDP.
    nondirectional_multipliers: dict
        Nondirectional components are sensitive to demands coming
        in any direction. Results are typically available in two
        orthogonal directions. FEMA P-58 suggests using the
        formula `max(dir_1, dir_2) * 1.2` to estimate the demand
        for such components. This parameter allows modifying the
        1.2 multiplier with a user-specified value. The change can
        be applied to "ALL" EDPs, or for specific EDPs, such as
        "PFA", "PFV", etc. Examples:
        #. {'PFA': 1.2, 'PID': 1.00}
        #. {'ALL': 1.0}
    demand_sample: pd.DataFrame
        Dataframe containing the demand sample, realizations of EDPs
        (or/and IMs) that are used for damage and loss calculations.

    Returns
    -------
    demand_dict: dict
        A dictionary of assembled demand data for calculation

    Raises
    ------
    ValueError
        If demand data for a given EDP cannot be found

    """
    demand_dict = {}

    for edps, expression in required_edps:
        edp_values = {}

        for edp in edps:
            edp_type, location, direction = edp.split('-')

            if direction == '0':
                # non-directional
                demand = (
                    demand_sample.loc[
                        :,  # type: ignore
                        (edp_type, location),
                    ]
                    .max(axis=1)
                    .to_numpy()
                )

                if edp_type in nondirectional_multipliers:
                    multiplier = nondirectional_multipliers[edp_type]

                elif 'ALL' in nondirectional_multipliers:
                    multiplier = nondirectional_multipliers['ALL']

                else:
                    msg = (
                        f'Peak orthogonal EDP multiplier '
                        f'for non-directional demand '
                        f'calculation of `{edp_type}` not specified.'
                    )
                    raise ValueError(msg)

                demand *= multiplier

            else:
                # directional
                demand = demand_sample[edp_type, location, direction].to_numpy()

            edp_values[edp] = demand

        # evaluate expression
        if expression is not None:
            # build a dict of values
            value_dict = {}
            for i, edp_value in enumerate(edp_values.values()):
                value_dict[f'X{i + 1}'] = edp_value
            demand = ne.evaluate(
                _clean_up_expression(expression), local_dict=value_dict
            )
        demand_dict[edps, expression] = demand

    return demand_dict


def _clean_up_expression(expression: str) -> str:
    """
    Clean up a mathematical expression in a string.

    Cleans up the given mathematical expression by ensuring it
    contains only allowed characters and replaces the caret (^)
    exponentiation operator with the double asterisk (**) operator.

    Parameters
    ----------
    expression: str
        The mathematical expression to clean up.

    Returns
    -------
    str
        The cleaned-up mathematical expression.

    Raises
    ------
    ValueError
        If the expression contains invalid characters.

    Examples
    --------
    >>> _clean_up_expression('3 + 5 * 2')
    '3 + 5 * 2'
    >>> _clean_up_expression('2^3')
    '2**3'
    >>> _clean_up_expression('2 ** 3')
    '2 ** 3'
    >>> _clean_up_expression(
    ...     "[o.fork() for (o,i) in "
    ...     "[(__import__('os'), __import__('itertools'))] "
    ...     "for x in i.repeat(0)]"
    ... )
    Traceback (most recent call last): ...

    """
    allowed_chars = re.compile(r'^[0-9a-zA-Z\^\+\-\*/\(\)\s]*$')
    if not bool(allowed_chars.match(expression)):
        msg = f'Invalid expression: {expression}'
        raise ValueError(msg)
    # replace exponantiation with `^` with the native `**` in case `^`
    # was used. But please use `**`..
    return expression.replace('^', '**')


def _verify_edps_available(available_edps: dict, required: set) -> None:
    """
    Verify EDP availability.

    Verifies that the required EDPs are available and raises
    appropriate errors otherwise.

    Parameters
    ----------
    available_edps: dict
        Dictionary mapping (`edp_type`-`cmp`-`dir`) to list of `loc`
        where values are available.
    required: set
        Set of required EDPs, expressed as
        `edp_type`-`loc`-`dir`. Direction `0` has special meaning: It
        is used for directional demands.

    Raises
    ------
    ValueError
        When the verification fails.

    """
    # Verify that the required EDPs are available in the
    # demand sample
    for edps, _ in required:
        for edp in edps:
            edp_type, location, direction = edp.split('-')
            if (edp_type, location) not in available_edps:
                msg = (
                    f'Unable to locate `{edp_type}` at location '
                    f'{location} in demand sample.'
                )
                raise ValueError(msg)
            # if non-directional demand is requested, ensure there
            # are entries (all directions accepted)
            num_entries = len(available_edps[edp_type, location])
            if edp[2] == '0' and num_entries == 0:
                msg = (
                    f'Unable to locate any `{edp_type}` '
                    f'at location {location} and direction {direction}.'
                )
                raise ValueError(msg)
            if edp[2] != '0' and num_entries == 0:
                msg = f'Unable to locate `{edp_type}-{location}-{direction}`.'
                raise ValueError(msg)



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# Pouria Kourehpaz
# Kuanshi Zhong
# John Vouvakis Manousakis

"""Classes and methods that handle file input and output."""

from __future__ import annotations

import json
from pathlib import Path

import numpy as np
import pandas as pd

from pelicun import base

convert_dv_name = {
    'DV_rec_cost': 'Reconstruction Cost',
    'DV_rec_time': 'Reconstruction Time',
    'DV_injuries_0': 'Injuries lvl. 1',
    'DV_injuries_1': 'Injuries lvl. 2',
    'DV_injuries_2': 'Injuries lvl. 3',
    'DV_injuries_3': 'Injuries lvl. 4',
    'DV_red_tag': 'Red Tag ',
}

dependency_to_acronym = {
    'btw. Fragility Groups': 'FG',
    'btw. Performance Groups': 'PG',
    'btw. Floors': 'LOC',
    'btw. Directions': 'DIR',
    'btw. Component Groups': 'CSG',
    'btw. Damage States': 'DS',
    'Independent': 'IND',
    'per ATC recommendation': 'ATC',
}

HAZUS_occ_converter = {
    'RES': 'Residential',
    'COM': 'Commercial',
    'REL': 'Commercial',
    'EDU': 'Educational',
    'IND': 'Industrial',
    'AGR': 'Industrial',
}

legacy_names = {
    'damage_DB_FEMA_P58_2nd': 'FEMA P-58',
    'damage_DB_Hazus_EQ_bldg': 'Hazus Earthquake - Buildings',
    'damage_DB_Hazus_EQ_story': 'Hazus Earthquake - Stories',
    'damage_DB_Hazus_EQ_trnsp': 'Hazus Earthquake - Transportation',
    'damage_DB_Hazus_EQ_water': 'Hazus Earthquake - Water',
    'damage_DB_Hazus_EQ_power': 'Hazus Earthquake - Power',
    'damage_DB_SimCenter_Hazus_HU_bldg': 'Hazus Hurricane Wind',
    'loss_repair_DB_FEMA_P58_2nd': 'FEMA P-58',
    'loss_repair_DB_Hazus_EQ_bldg': 'Hazus Earthquake - Buildings',
    'loss_repair_DB_Hazus_EQ_story': 'Hazus Earthquake - Stories',
    'loss_repair_DB_Hazus_EQ_trnsp': 'Hazus Earthquake - Transportation',
    'loss_repair_DB_SimCenter_Hazus_HU_bldg': 'Hazus Hurricane Wind',
}


def save_to_csv(  # noqa: C901
    data: pd.DataFrame | None,
    filepath: Path | None,
    units: pd.Series | None = None,
    unit_conversion_factors: dict | None = None,
    orientation: int = 0,
    *,
    use_simpleindex: bool = True,
    log: base.Logger | None = None,
) -> pd.DataFrame | None:
    """
    Save data to a CSV file following the standard SimCenter schema.

    The produced CSV files have a single header line and an index
    column. The second line may start with 'Units' in the index or the
    first column may be 'Units' to provide the units for the data in
    the file.

    Parameters
    ----------
    data: DataFrame
        The data to save.
    filepath: Path
        The location of the destination file. If None, the data is not
        saved, but returned in the end.
    units: Series, optional
        Provides a Series with variables and corresponding units.
    unit_conversion_factors: dict, optional
        Dictionary containing key-value pairs of unit names and their
        corresponding factors. Conversion factors are defined as the
        number of times a base unit fits in the alternative unit.
    orientation: int, {0, 1}, default 0
        If 0, variables are organized along columns; otherwise, they
        are along the rows. This is important when converting values
        to follow the prescribed units.
    use_simpleindex: bool, default True
        If True, MultiIndex columns and indexes are converted to
        SimpleIndex before saving.
    log: Logger, optional
        Logger object to be used. If no object is specified, no
        logging is performed.

    Raises
    ------
    ValueError
        If units is not None but unit_conversion_factors is None.
    ValueError
        If writing to a file fails.
    ValueError
        If the provided file name does not have the `.csv` suffix.

    Returns
    -------
    DataFrame or None
        If `filepath` is None, returns the DataFrame with potential
        unit conversions and reformatting applied. Otherwise, returns
        None after saving the data to a CSV file.

    """
    if filepath is None:
        if log:
            log.msg('Preparing data ...', prepend_timestamp=False)

    elif log:
        log.msg(f'Saving data to `{filepath!s}`...', prepend_timestamp=False)

    if data is None:
        if log:
            log.warning('Data was empty, no file saved.')
        return None

    assert isinstance(data, pd.DataFrame)

    # make sure we do not modify the original data
    data = data.copy()

    # convert units and add unit information, if needed
    if units is not None:
        if unit_conversion_factors is None:
            msg = (
                'When `units` is not None, '
                '`unit_conversion_factors` must be provided.'
            )
            raise ValueError(msg)

        if log:
            log.msg('Converting units...', prepend_timestamp=False)

        # if the orientation is 1, we might not need to scale all columns
        if orientation == 1:
            cols_to_scale_bool = [dt in {float, int} for dt in data.dtypes]
            cols_to_scale = data.columns[cols_to_scale_bool]

        labels_to_keep = []

        for unit_name in units.unique():
            labels = units.loc[units == unit_name].index.to_numpy()

            unit_factor = 1.0 / unit_conversion_factors[unit_name]

            active_labels = []

            if orientation == 0:
                for label in labels:
                    if label in data.columns:
                        active_labels.append(label)  # noqa: PERF401

                if len(active_labels) > 0:
                    data.loc[:, active_labels] *= unit_factor

            else:  # elif orientation == 1:
                for label in labels:
                    if label in data.index:
                        active_labels.append(label)  # noqa: PERF401

                if len(active_labels) > 0:
                    data.loc[np.array(active_labels), np.array(cols_to_scale)] *= (
                        unit_factor
                    )

            labels_to_keep += active_labels

        units_df = units.loc[labels_to_keep].to_frame()

        if orientation == 0:
            data = pd.concat([units_df.T, data], axis=0)
            data = data.sort_index(axis=1)
        else:
            data = pd.concat([units_df, data], axis=1)
            data = data.sort_index()

        if log:
            log.msg('Unit conversion successful.', prepend_timestamp=False)

    assert isinstance(data, pd.DataFrame)
    if use_simpleindex:
        # convert MultiIndex to regular index with '-' separators
        if isinstance(data.index, pd.MultiIndex):
            data = base.convert_to_SimpleIndex(data)

        # same thing for the columns
        if isinstance(data.columns, pd.MultiIndex):
            data = base.convert_to_SimpleIndex(data, axis=1)

    if filepath is not None:
        if filepath.suffix == '.csv':
            # save the contents of the DataFrame into a csv
            data.to_csv(filepath)

            if log:
                log.msg('Data successfully saved to file.', prepend_timestamp=False)

        else:
            msg = (
                f'Please use the `.csv` file extension. '
                f'Received file name is `{filepath}`'
            )
            raise ValueError(msg)

        return None

    # at this line, filepath is None
    return data


def substitute_default_path(
    data_paths: list[str | pd.DataFrame], log: base.Logger | None = None
) -> list[str | pd.DataFrame]:
    """
    Substitute the default directory path.

    This function iterates over a list of data paths and replaces
    those with the 'PelicunDefault/' substring with the full paths to
    model files in the built-in Damage and Loss Model Library.
    Default paths are expected to follow the
    `PelicunDefault/method_name/model_type.extension` structure. The
    `method_name` identifies the methodology from those available in the
    `{base.pelicun_path}/resources/dlml_resource_paths.json` file. The
    `model_type` identifies the type of model requested. Currently, the
    following types are supported: 'fragility', 'consequence_repair',
    'loss_repair'. The `extension` is intended to identify 'CSV' files with
    model parameters and 'JSON' files with metadata.
    The `model_type` and `extension` strings are not limited to the
    supported values. If you know a particular file exists in the method's
    folder, you can use the corresponding `model_type.extension` to access
    that file.

    Parameters
    ----------
    data_paths: list of str or pd.DataFrame
        A list containing the paths to data files. These paths may
        include a placeholder directory 'PelicunDefault/' that needs
        to be substituted with the actual path specified in the
        resource mapping.
    log: Logger
        Logger object to be used. If no object is specified, no logging
        is performed.

    Returns
    -------
    list of str or pd.DataFrame

    Raises
    ------
    KeyError
      If the method_name after 'PelicunDefault/' does not exist in the
      `resource_paths` keys.
      If the method_name after 'PelicunDefault/' does not exist in the
      legacy list of filenames preserved for backwards compatibility.

    Notes
    -----
    - The function assumes that `base.pelicun_path` is properly
      initialized and points to the correct directory where resources
      are located.
    - If a path in the input list does not contain 'PelicunDefault/',
      the path is added to the output list unchanged.

    Examples
    --------
    >>> data_paths = ['PelicunDefault/Hazus Hurricane/fragility.csv', 'data/file2.txt']
    >>> substitute_default_path(data_paths)
    ['{base.pelicun_path}/resources/DamageAndLossModelLibrary/'
      'hurricane/building/portfolio/Hazus v5.1 coupled/fragility.csv',
      'data/file2.txt']

    """
    # Load the resource paths from the JSON file
    resource_file_path = (
        Path(base.pelicun_path) / 'resources' / 'dlml_resource_paths.json'
    )
    with resource_file_path.open('r') as file:
        resource_paths = json.load(file)

    updated_paths: list[str | pd.DataFrame] = []
    for data_path_str in data_paths:
        if isinstance(data_path_str, str) and 'PelicunDefault/' in data_path_str:
            data_path = Path(data_path_str)
            # Extract the filename from the end after 'PelicunDefault/'
            file_name = data_path.parts[-1]

            # Check if there is a method name identified
            method_name = data_path.parts[-2]

            # <backwards compatibility>
            if method_name == 'PelicunDefault':
                # No method name, check for legacy input
                if file_name.startswith(
                    ('fragility_DB', 'damage_DB', 'bldg_repair_DB', 'loss_repair_DB')
                ):
                    if log:
                        log.warning(
                            'Default libraries are no longer referenced using '
                            'the following placeholder filenames after "PelicunDB/": '
                            '`fragility_DB...`, `damage_DB...`, `bldg_repair_DB...`, '
                            '`loss_repair_DB...`. Such inputs will lead to errors in '
                            'future versions of pelicun. Please replace such '
                            'references with a combination of a specific method and '
                            'data type. For example, use '
                            '`PelicunDefault/FEMA P-58/fragility` to get FEMA P-58 '
                            'damage models, and '
                            '`PelicunDefault/Hazus Hurricane/consequence_repair` to '
                            'get Hazus hurricane consequence models. See the online '
                            'documentation for more details.'
                        )

                    method_name = legacy_names[file_name.split('.')[0]]
                    if file_name.startswith(('fragility', 'damage')):
                        data_type = 'fragility'
                    else:
                        data_type = 'consequence_repair'

                    extension = file_name.split('.')[-1]
                    file_name = f'{data_type}.{extension}'

                else:
                    msg = f'Default data path `{data_path_str}` not recognized.'
                    raise KeyError(msg)

            # Check if the method name exists in the resource paths dictionary
            if method_name not in resource_paths:
                msg = f'Method `{method_name}` not found in resource paths.'
                raise KeyError(msg)
            method_folder = resource_paths[method_name]

            # Substitute the default path with a full path to the file
            updated_path = str(
                Path(base.pelicun_path)
                / 'resources'
                / 'DamageAndLossModelLibrary'
                / method_folder
                / file_name
            )
            updated_paths.append(updated_path)
        else:
            updated_paths.append(data_path_str)

    return updated_paths


def load_data(  # noqa: C901
    data_source: str | pd.DataFrame,
    unit_conversion_factors: dict | None = None,
    orientation: int = 0,
    *,
    reindex: bool = True,
    return_units: bool = False,
    log: base.Logger | None = None,
) -> tuple[pd.DataFrame, pd.Series] | pd.DataFrame:
    """
    Load data assuming it follows standard SimCenter tabular schema.

    The data is assumed to have a single header line and an index column. The
    second line may start with 'Units' in the index and provide the units for
    each column in the file.

    Parameters
    ----------
    data_source: string or DataFrame
        If it is a string, the data_source is assumed to point to the location
        of the source file. If it is a DataFrame, the data_source is assumed to
        hold the raw data.
    unit_conversion_factors: dict, optional
        Dictionary containing key-value pairs of unit names and their
        corresponding factors. Conversion factors are defined as the
        number of times a base unit fits in the alternative unit. If
        no conversion factors are specified, then no unit conversions
        are made.
    orientation: int, {0, 1}, default: 0
        If 0, variables are organized along columns; otherwise they are along
        the rows. This is important when converting values to follow the
        prescribed units.
    reindex: bool
        If True, reindexes the table to ensure a 0-based, continuous index
    return_units: bool
        If True, returns the units as well as the data to allow for adjustments
        in unit conversion.
    log: Logger
        Logger object to be used. If no object is specified, no logging
        is performed.

    Returns
    -------
    tuple
        data: DataFrame
            Parsed data.
        units: Series
            Labels from the data and corresponding units specified in the
            data. Units are only returned if return_units is set to True.

    Raises
    ------
    TypeError
        If `data_source` is neither a string nor a DataFrame, a
        TypeError is raised.

    """
    if isinstance(data_source, pd.DataFrame):
        # store it at proceed (copying is needed to avoid changing the
        # original)
        data = base.with_parsed_str_na_values(data_source.copy())
    elif isinstance(data_source, str):
        # otherwise, load the data from a file
        data = load_from_file(data_source)
    else:
        msg = f'Invalid data_source type: {type(data_source)}'
        raise TypeError(msg)

    # Define a dictionary to decide the axis based on the orientation
    axis = {0: 1, 1: 0}
    the_index = data.columns if orientation == 1 else data.index

    # Check for units information (case-insensitive)
    # if there is information about units, separate that information
    # and optionally apply conversions to all numeric values
    units_key = None
    for key in the_index:
        if str(key).lower() == 'units':
            units_key = key
            break

    if units_key is not None:
        units = data[units_key] if orientation == 1 else data.loc[units_key]
        data = data.drop([units_key], axis=orientation)  # type: ignore
        data = base.convert_dtypes(data)

        if unit_conversion_factors is not None:
            numeric_elements = (
                (data.select_dtypes(include=[np.number]).index)  # type: ignore
                if orientation == 0
                else (
                    data.select_dtypes(include=[np.number]).columns  # type: ignore
                )
            )

            if log:
                log.msg('Converting units...', prepend_timestamp=False)

            conversion_factors = units.map(
                lambda unit: (
                    1.00
                    if pd.isna(unit)
                    else unit_conversion_factors.get(unit, 1.00)
                )
            )

            if orientation == 1:
                data.loc[:, numeric_elements] = data.loc[
                    :, numeric_elements  # type: ignore
                ].multiply(
                    conversion_factors,
                    axis=axis[orientation],  # type: ignore
                )  # type: ignore
            else:
                data.loc[numeric_elements, :] = data.loc[
                    numeric_elements, :
                ].multiply(
                    conversion_factors,
                    axis=axis[orientation],  # type: ignore
                )  # type: ignore

        if log:
            log.msg('Unit conversion successful.', prepend_timestamp=False)

    else:
        units = None
        data = base.convert_dtypes(data)

    # convert columns or index to MultiIndex if needed
    data = base.convert_to_MultiIndex(data, axis=1)
    data = data.sort_index(axis=1)

    # reindex the data, if needed
    if reindex:
        data.index = pd.RangeIndex(start=0, stop=data.shape[0], step=1)
    else:
        # convert index to MultiIndex if needed
        data = base.convert_to_MultiIndex(data, axis=0)
        data = data.sort_index()

    if return_units:
        if units is not None:
            # convert index in units Series to MultiIndex if needed
            units = base.convert_to_MultiIndex(units, axis=0).dropna()  # type: ignore
            units = units.sort_index()
        output = data, units
    else:
        output = data  # type: ignore

    return output  # type: ignore


def load_from_file(filepath: str, log: base.Logger | None = None) -> pd.DataFrame:
    """
    Load data from a file and stores it in a DataFrame.

    Currently, only CSV files are supported, but the function is easily
    extensible to support other file formats.

    Parameters
    ----------
    filepath: string
        The location of the source file.
    log: base.Logger, optional
        Optional logger object.

    Returns
    -------
    tuple
        data: DataFrame
            Data loaded from the file.
        log: Logger
            Logger object to be used. If no object is specified, no logging
            is performed.

    Raises
    ------
    FileNotFoundError
        If the filepath is invalid.
    ValueError
        If the file is not a CSV.

    """
    if log:
        log.msg(f'Loading data from {filepath}...')

    # check if the filepath is valid
    filepath_path = Path(filepath).resolve()

    if not filepath_path.is_file():
        msg = (
            f'The filepath provided does not point to an existing '
            f'file: {filepath_path}'
        )
        raise FileNotFoundError(msg)

    if filepath_path.suffix == '.csv':
        # load the contents of the csv into a DataFrame

        data = pd.read_csv(
            filepath_path,
            header=0,
            index_col=0,
            low_memory=False,
            encoding_errors='replace',
        )

        if log:
            log.msg('File successfully opened.', prepend_timestamp=False)

    else:
        msg = (
            f'Unexpected file type received when trying '
            f'to load from csv: {filepath_path}'
        )
        raise ValueError(msg)

    return data



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.
#
# Contributors:
# Adam Zsarnóczay
# John Vouvakis Manousakis

"""These are unit and integration tests on the auto module of pelicun."""

from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from pelicun.auto import auto_populate

# The tests maintain the order of definitions of the `auto.py` file.

#  _____                 _   _
# |  ___|   _ _ __   ___| |_(_) ___  _ __  ___
# | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
# |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
#
# The following tests verify the functions of the module.


@pytest.fixture
def setup_valid_config() -> dict:
    return {'GeneralInformation': {'someKey': 'someValue'}}


@pytest.fixture
def setup_auto_script_path() -> str:
    return 'PelicunDefault/test_script'


@pytest.fixture
def setup_expected_base_path() -> str:
    return '/expected/path/resources/auto/'


"""
def test_valid_inputs(setup_valid_config: dict, setup_auto_script_path: str) -> None:
    with patch('pelicun.base.pelicun_path', '/expected/path'), patch(
        'os.path.exists', return_value=True
    ), patch('importlib.__import__') as mock_import:
        mock_auto_populate_ext = MagicMock(
            return_value=({'AIM_ap': 'value'}, {'DL_ap': 'value'}, 'CMP')
        )
        mock_import.return_value.auto_populate = mock_auto_populate_ext

        config, cmp = auto_populate(setup_valid_config, Path(setup_auto_script_path))

        assert 'DL' in config
        assert cmp == 'CMP'
"""


def test_missing_general_information() -> None:
    with pytest.raises(
        ValueError,
        match='No Asset Information provided for the auto-population routine.',
    ):
        auto_populate({}, Path('some/path'))


def test_pelicun_default_path_replacement(
    setup_auto_script_path: str, setup_expected_base_path: str
) -> None:
    modified_path = setup_auto_script_path.replace(
        'PelicunDefault/', setup_expected_base_path
    )
    assert modified_path.startswith(setup_expected_base_path)


"""
def test_auto_population_script_execution(
    setup_valid_config: dict, setup_auto_script_path: str
) -> None:
    with patch('pelicun.base.pelicun_path', '/expected/path'), patch(
        'os.path.exists', return_value=True
    ), patch('importlib.__import__') as mock_import:
        mock_auto_populate_ext = MagicMock(
            return_value=({'AIM_ap': 'value'}, {'DL_ap': 'value'}, 'CMP')
        )
        mock_import.return_value.auto_populate = mock_auto_populate_ext

        auto_populate(setup_valid_config, Path(setup_auto_script_path))
        mock_import.assert_called_once()
"""



#
# Copyright (c) 2018 Leland Stanford Junior University
# Copyright (c) 2018 The Regents of the University of California
#
# This file is part of pelicun.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# You should have received a copy of the BSD 3-Clause License along with
# pelicun. If not, see <http://www.opensource.org/licenses/>.

"""
Validation test on loss functions.

In this example, a single loss function is defined as a 1:1 mapping of
the input EDP.  This means that the resulting loss distribution will
be the same as the EDP distribution, allowing us to test and confirm
that this is what happens.

"""

from __future__ import annotations

import numpy as np
import pandas as pd

from pelicun import assessment, file_io


def test_validation_loss_function() -> None:
    sample_size = 100000

    # initialize a pelicun assessment
    asmnt = assessment.Assessment({'PrintLog': False, 'Seed': 42})

    #
    # Demands
    #

    demands = pd.DataFrame(
        {
            'Theta_0': [0.50],
            'Theta_1': [0.90],
            'Family': ['lognormal'],
            'Units': ['mps2'],
        },
        index=pd.MultiIndex.from_tuples(
            [
                ('PFA', '0', '1'),
            ],
        ),
    )

    asmnt.demand.load_model({'marginals': demands})

    asmnt.demand.generate_sample({'SampleSize': sample_size})

    #
    # Asset
    #

    asmnt.stories = 1

    cmp_marginals = pd.read_csv(
        'pelicun/tests/validation/v0/data/CMP_marginals.csv', index_col=0
    )
    cmp_marginals['Blocks'] = cmp_marginals['Blocks']
    asmnt.asset.load_cmp_model({'marginals': cmp_marginals})

    asmnt.asset.generate_cmp_sample(sample_size)

    #
    # Damage
    #

    # nothing to do here.

    #
    # Losses
    #

    asmnt.loss.decision_variables = ('Cost',)

    loss_map = pd.DataFrame(['cmp.A'], columns=['Repair'], index=['cmp.A'])
    asmnt.loss.add_loss_map(loss_map)

    loss_functions = file_io.load_data(
        'pelicun/tests/validation/v0/data/loss_functions.csv',
        reindex=False,
        unit_conversion_factors=asmnt.unit_conversion_factors,
    )
    assert isinstance(loss_functions, pd.DataFrame)
    asmnt.loss.load_model_parameters([loss_functions])
    asmnt.loss.calculate()

    loss, _ = asmnt.loss.aggregate_losses(future=True)
    assert isinstance(loss, pd.DataFrame)

    loss_vals = loss['repair_cost'].to_numpy()

    # sample median should be close to 0.05
    assert np.allclose(np.median(loss_vals), 0.05, atol=1e-2)
    # dispersion should be close to 0.9
    assert np.allclose(np.log(loss_vals).std(), 0.90, atol=1e-2)

    # TODO(JVM): also test save/load sample
    # asmnt.loss.save_sample('/tmp/sample.csv')
    # asmnt.loss.load_sample('/tmp/sample.csv')
